from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.db.models.deletion import ProtectedError

from edc_dashboard.view_mixins import EdcViewMixin
from edc_lab import PACKED
from edc_lab import Manifest as ManifestObject
from edc_lab.labels import BoxLabel
from edc_lab.models import Box, Manifest
from edc_label import LabPrintersMixin, add_job_results_to_messages

from .action_view import ActionView


class PackView(EdcViewMixin, LabPrintersMixin, ActionView):
    post_action_url = "pack_listboard_url"
    valid_form_actions = [
        "add_selected_to_manifest",
        "remove_selected_items",
        "print_labels",
    ]
    label_cls = BoxLabel

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._selected_manifest = None

    def process_form_action(self, request=None):
        if not self.selected_items:
            message = "Nothing to do. No items have been selected."
            messages.warning(request, message)
        elif self.action == "remove_selected_items":
            self.remove_selected_items()
        elif self.action == "add_selected_to_manifest":
            if self.selected_manifest:
                self.add_selected_to_manifest()
        elif self.action == "print_labels":
            job_result = self.print_labels(pks=self.selected_items, request=request)
            if job_result:
                add_job_results_to_messages(request, [job_result])

    @property
    def selected_manifest(self):
        if not self._selected_manifest:
            if self.request.POST.get("selected_manifest"):
                try:
                    self._selected_manifest = Manifest.objects.get(
                        pk=self.request.POST.get("selected_manifest")
                    )
                except ObjectDoesNotExist:
                    pass
        return self._selected_manifest

    def add_selected_to_manifest(self):
        """Adds the selected items to the selected manifest."""
        if not self.selected_items:
            message = "Nothing to do. No items have been selected."
            messages.warning(self.request, message)
        elif not self.selected_manifest:
            message = "Nothing to do. No manifest has been selected."
            messages.warning(self.request, message)
        else:
            manifest_object = ManifestObject(
                manifest=self.selected_manifest, request=self.request
            )
            try:
                added = 0
                for selected_item in self.selected_items:
                    box = Box.objects.get(pk=selected_item)
                    if manifest_object.add_box(
                        box=box, manifest_item_identifier=box.box_identifier
                    ):
                        added += 1
                        box.status = PACKED
                        box.save()
                    else:
                        break
                if added > 0:
                    message = f"{added} items have been added to manifest {self.selected_manifest.human_readable_identifier}."
                    messages.success(self.request, message)
            except ProtectedError:
                message = "Unable to remove. Box is not empty."
                messages.error(self.request, message)

    def remove_selected_items(self):
        """Deletes the selected boxes, if allowed."""
        if not self.selected_items:
            message = "Nothing to do. No items have been selected."
            messages.warning(self.request, message)
        else:
            try:
                deleted = Box.objects.filter(pk__in=self.selected_items).delete()
                message = f"{deleted[0]} items have been removed."
                messages.success(self.request, message)
            except ProtectedError:
                message = "Unable to remove. Box is not empty."
                messages.error(self.request, message)



# Generated by Django 2.1 on 2018-10-22 05:54

import _socket
import django_audit_fields.fields.uuid_auto_field
import django_revision.revision_field
from django.db import migrations, models

import edc_model_fields.fields.hostname_modification_field
import edc_model_fields.fields.userfield
import django.utils.timezone


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Notification",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(max_length=25, unique=True)),
                ("display_name", models.CharField(max_length=255, unique=True)),
                ("enabled", models.BooleanField(default=True)),
            ],
            options={"ordering": ("display_name",)},
        )
    ]



from django.db import models

from edc_constants.choices import YES_NO
from edc_constants.constants import YES


class CaretakerFieldsMixin(models.Model):
    """A fields mixin for visit models where information
    on the participant is offered by another person,
    as in the case of infant and mother.

    One the ModelForm, override the default form to customize
    the choices and labels.

    """

    information_provider = models.CharField(
        verbose_name=(
            "Please indicate who provided most of the information for this participant's visit"
        ),
        max_length=20,
    )

    information_provider_other = models.CharField(
        verbose_name="if information provider is Other, please specify",
        max_length=20,
        blank=True,
        default="",
    )

    is_present = models.CharField(
        max_length=10,
        verbose_name="Is the participant present at today's visit",
        choices=YES_NO,
        default=YES,
    )

    class Meta:
        abstract = True



from __future__ import annotations

from decimal import Decimal
from typing import TYPE_CHECKING

from django.apps import apps as django_apps

from .get_unit_qty_out import get_unit_qty_out

if TYPE_CHECKING:
    from ..models import Allocation, Stock, StockAdjustment


def get_allocation_model_cls() -> Allocation:
    return django_apps.get_model("edc_pharmacy.allocation")


def get_stockadjustment_model_cls() -> StockAdjustment:
    return django_apps.get_model("edc_pharmacy.stockadjustment")


def update_stock_instance(stock: Stock) -> Stock:
    """Update stock instance fields 'unit_qty_in', 'unit_qty_out',
    'qty', 'qty_out'
    """
    # check if unit_qty_in has been manually adjusted or not
    if get_stockadjustment_model_cls().objects.filter(stock=stock).exists():
        # use adjusted value
        stock.unit_qty_in = (
            get_stockadjustment_model_cls()
            .objects.filter(stock=stock)
            .order_by("adjustment_datetime")
            .last()
            .unit_qty_in_new
        )
    else:
        # confirm default value from container definition
        stock.unit_qty_in = Decimal(stock.qty_in) * Decimal(stock.container.qty)
    # recalculate unit_qty_out
    stock.unit_qty_out = get_unit_qty_out(stock)
    # update overall container quantity if in-out==0
    if stock.unit_qty_out == stock.unit_qty_in:
        stock.qty_out = 1
        stock.qty = 0
    stock.save(update_fields=["unit_qty_in", "unit_qty_out", "qty", "qty_out"])
    stock.refresh_from_db()
    return stock



# -*- coding: utf-8 -*-
# Generated by Django 1.10.3 on 2016-11-27 22:26
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [("edc_appointment", "0002_auto_20161126_1156")]

    operations = [
        migrations.AlterModelOptions(name="appointment", options={}),
        migrations.RemoveField(model_name="historicalappointment", name="consent_version"),
        migrations.AlterField(
            model_name="appointment",
            name="created",
            field=models.DateTimeField(default=django.utils.timezone.now, editable=False),
        ),
        migrations.AlterField(
            model_name="appointment",
            name="modified",
            field=models.DateTimeField(default=django.utils.timezone.now, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="created",
            field=models.DateTimeField(default=django.utils.timezone.now, editable=False),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="modified",
            field=models.DateTimeField(default=django.utils.timezone.now, editable=False),
        ),
        migrations.AlterUniqueTogether(name="appointment", unique_together=set([])),
        migrations.RemoveField(model_name="appointment", name="consent_version"),
    ]



# Generated by Django 6.0 on 2025-03-12 11:33

from django.db import migrations
from tqdm import tqdm

from ..models import (
    Allocation,
    Dispense,
    DispenseItem,
    StockTransferItem,
    StorageBin,
    StorageBinItem,
)


def update_site_id(apps, schema_editor):

    stock_transfer_confirmation = apps.get_model("edc_pharmacy", "StockTransferConfirmation")
    stock_transfer_confirmation_item = apps.get_model(
        "edc_pharmacy", "StockTransferConfirmationItem"
    )
    for model_cls in [
        Allocation,
        stock_transfer_confirmation,
        stock_transfer_confirmation_item,
        StorageBin,
        StorageBinItem,
        Dispense,
        DispenseItem,
    ]:
        total = model_cls.objects.count()
        for obj in tqdm(model_cls.objects.all(), total=total):
            obj.save(update_fields=("site",))

    for obj in tqdm(Allocation.objects.all(), total=Allocation.objects.count()):
        stock = obj.stock
        stock.allocated = True
        stock.save(update_fields=("allocated",))

    for obj in tqdm(StockTransferItem.objects.all(), total=StockTransferItem.objects.count()):
        stock = obj.stock
        stock.transferred = True
        stock.save(update_fields=("transferred",))


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_pharmacy",
            "0061_alter_historicalstocktransferconfirmation_options_and_more",
        ),
    ]

    operations = [migrations.RunPython(update_site_id)]



from __future__ import annotations

from django.core.exceptions import ObjectDoesNotExist

from edc_utils.round_up import round_half_away_from_zero

from ..exceptions import ConversionNotHandled
from ..units import (
    GRAMS_PER_LITER,
    MICROMOLES_PER_LITER,
    MILLIGRAMS_PER_DECILITER,
    MILLIGRAMS_PER_LITER,
    MILLIMOLES_PER_LITER,
)
from .molecular_weight_model_cls import molecular_weight_model_cls

__all__ = ["convert_units"]


def get_mw(label):
    try:
        molecular_weight = molecular_weight_model_cls().objects.get(label=label)
    except ObjectDoesNotExist as e:
        raise ConversionNotHandled(
            f"Conversion not handled. Molecular weight not found for {label}."
        ) from e
    else:
        mw = molecular_weight.mw
    return mw


def micromoles_per_liter_to(*, label: str, value: float | int, units_to: str) -> dict:
    if units_to == MICROMOLES_PER_LITER:
        return {MICROMOLES_PER_LITER: float(value)}
    if units_to == MILLIMOLES_PER_LITER:
        return {MILLIMOLES_PER_LITER: float(value) / 1000.00}
    if units_to == GRAMS_PER_LITER:
        return {GRAMS_PER_LITER: (float(value) * get_mw(label)) / 100.00}
    if units_to == MILLIGRAMS_PER_DECILITER:
        return {MILLIGRAMS_PER_DECILITER: (float(value) * get_mw(label)) / 10000.00}
    raise ConversionNotHandled(f"Conversion not found. Tried umol/L to {units_to}. ")


def milligrams_per_deciliter_to(
    *, label: str, value: float | int, units_to: str
) -> dict[str, float]:
    if units_to == MILLIGRAMS_PER_DECILITER:
        return {MILLIGRAMS_PER_DECILITER: float(value)}
    if units_to == MILLIMOLES_PER_LITER:
        return {MILLIMOLES_PER_LITER: (float(value) * 10.00) / get_mw(label)}
    if units_to == MICROMOLES_PER_LITER:
        return {MICROMOLES_PER_LITER: (float(value) * 10000.00) / get_mw(label)}
    if units_to == MILLIGRAMS_PER_LITER:
        return {MILLIGRAMS_PER_LITER: float(value) * 10.00}
    if units_to == GRAMS_PER_LITER:
        return {GRAMS_PER_LITER: float(value) / 100.00}
    raise ConversionNotHandled(f"Conversion not found. Tried mg/dL to {units_to}. ")


class UnitsConverter:
    def __init__(
        self,
        *,
        label: str,
        value: int | float,
        units_from: str,
        units_to: str,
        places: int | None = None,
    ):
        self.label = label
        self.value = value
        self.units_from = units_from
        self.units_to = units_to
        self.places = places or 4

        if label is None:
            raise ValueError("label is required. See convert_units.")
        if value is not None and units_from and units_to and units_from != units_to:
            self.converted_value = self.get_converted_value()
        elif units_from == units_to:
            self.converted_value = value
        else:
            raise ConversionNotHandled(
                f"Conversion not handled. Tried {label} from {units_from} to {units_to}."
            )

    def get_mw(self):
        return get_mw(self.label)

    def round_up(self, converted_value):
        try:
            converted_value = round_half_away_from_zero(converted_value, self.places)
        except TypeError as e:
            raise ConversionNotHandled(
                f"Conversion not handled. Tried {self.label} from {self.units_from} "
                f"to {self.units_to}. "
                f"Got {e} when rounding {converted_value} to {self.places} places."
            ) from e
        return converted_value

    def from_milligrams_per_deciliter(self) -> float | int:
        if self.units_to != MILLIGRAMS_PER_DECILITER:
            return milligrams_per_deciliter_to(
                label=self.label, value=float(self.value), units_to=self.units_to
            )[self.units_to]
        return self.value

    def from_grams_per_liter(self) -> float | int:
        if self.units_to != GRAMS_PER_LITER:
            return milligrams_per_deciliter_to(
                label=self.label,
                value=float(self.value) * 100.00,
                units_to=self.units_to,
            )[self.units_to]
        return self.value

    def from_millimoles_per_liter(self) -> float | int:
        if self.units_to != MILLIMOLES_PER_LITER:
            return micromoles_per_liter_to(
                label=self.label, value=self.value / 1000, units_to=self.units_to
            )[self.units_to]
        return self.value

    def from_micromoles_per_liter(self):
        if self.units_to != MICROMOLES_PER_LITER:
            return micromoles_per_liter_to(
                label=self.label, value=self.value, units_to=self.units_to
            )[self.units_to]
        return self.value

    def get_converted_value(self) -> int | float:
        converted_value = None
        if self.units_from == MILLIGRAMS_PER_DECILITER:
            converted_value = self.from_milligrams_per_deciliter()
        elif self.units_from == GRAMS_PER_LITER:
            converted_value = self.from_grams_per_liter()
        elif self.units_from == MILLIMOLES_PER_LITER:
            converted_value = self.from_millimoles_per_liter()
        elif self.units_from == MICROMOLES_PER_LITER:
            converted_value = self.from_micromoles_per_liter()
        if not converted_value:
            raise ConversionNotHandled(
                f"Conversion not handled. Tried {self.label} from "
                f"{self.units_from} to {self.units_to}."
            )
        return self.round_up(converted_value)


def convert_units(
    *,
    label: str | None = None,
    value: int | float | None = None,
    units_from: str | None = None,
    units_to: str | None = None,
    places: int | None = None,
):
    return UnitsConverter(
        label=label,
        value=value,
        units_from=units_from,
        units_to=units_to,
        places=places,
    ).converted_value



from django.apps import apps as django_apps
from django.conf import settings

from edc_action_item.action import Action
from edc_action_item.site_action_items import site_action_items
from edc_adverse_event.constants import AE_INITIAL_ACTION
from edc_constants.constants import HIGH_PRIORITY, YES
from edc_visit_schedule.utils import is_baseline

from .constants import (
    BLOOD_RESULTS_EGFR_ACTION,
    BLOOD_RESULTS_FBC_ACTION,
    BLOOD_RESULTS_GLU_ACTION,
    BLOOD_RESULTS_HBA1C_ACTION,
    BLOOD_RESULTS_INSULIN_ACTION,
    BLOOD_RESULTS_LFT_ACTION,
    BLOOD_RESULTS_LIPIDS_ACTION,
    BLOOD_RESULTS_RFT_ACTION,
)

subject_app_label = getattr(
    settings, "EDC_BLOOD_RESULTS_MODEL_APP_LABEL", settings.SUBJECT_APP_LABEL
)


class BaseResultsAction(Action):
    name = None
    display_name = None
    reference_model = None
    priority = HIGH_PRIORITY
    show_on_dashboard = True
    create_by_user = False

    def reopen_action_item_on_change(self):
        return False

    def get_next_actions(self):
        next_actions = []
        if self.is_reportable and not is_baseline(instance=self.reference_obj.related_visit):
            # AE for reportable result, though at baseline
            next_actions = [AE_INITIAL_ACTION]
        return next_actions

    @property
    def is_reportable(self):
        return (
            self.reference_obj.results_abnormal == YES
            and self.reference_obj.results_reportable == YES
        )


class BloodResultsLftAction(BaseResultsAction):
    name = BLOOD_RESULTS_LFT_ACTION
    display_name = "Reportable result: LFT"
    reference_model = f"{subject_app_label}.bloodresultslft"


class BloodResultsRftAction(BaseResultsAction):
    name = BLOOD_RESULTS_RFT_ACTION
    display_name = "Reportable result: RFT"
    reference_model = f"{subject_app_label}.bloodresultsrft"


class BloodResultsFbcAction(BaseResultsAction):
    name = BLOOD_RESULTS_FBC_ACTION
    display_name = "Reportable result: FBC"
    reference_model = f"{subject_app_label}.bloodresultsfbc"


class BloodResultsLipidsAction(BaseResultsAction):
    name = BLOOD_RESULTS_LIPIDS_ACTION
    display_name = "Reportable result: LIPIDS"
    reference_model = f"{subject_app_label}.bloodresultslipids"


class BloodResultsEgfrAction(BaseResultsAction):
    name = BLOOD_RESULTS_EGFR_ACTION
    display_name = "Reportable eGFR"
    reference_model = f"{subject_app_label}.bloodresultsrft"

    def get_next_actions(self):
        next_actions = []
        if (
            not is_baseline(instance=self.reference_obj.related_visit)
            and self.reference_obj.egfr_value < 45
        ):
            # AE for reportable result, though not on DAY1.0
            next_actions = [AE_INITIAL_ACTION]
        return next_actions


class BloodResultsGluAction(BaseResultsAction):
    name = BLOOD_RESULTS_GLU_ACTION
    display_name = "Reportable Blood Glucose"
    reference_model = f"{subject_app_label}.bloodresultsglu"


class BloodResultsHba1cAction(BaseResultsAction):
    name = BLOOD_RESULTS_HBA1C_ACTION
    display_name = "Reportable HbA1c"
    reference_model = f"{subject_app_label}.bloodresultshba1c"


class BloodResultsInsulinAction(BaseResultsAction):
    name = BLOOD_RESULTS_INSULIN_ACTION
    display_name = "Reportable Insulin"
    reference_model = f"{subject_app_label}.bloodresultsins"


def register_actions():
    for action_cls in [
        BloodResultsEgfrAction,
        BloodResultsFbcAction,
        BloodResultsGluAction,
        BloodResultsHba1cAction,
        BloodResultsInsulinAction,
        BloodResultsLftAction,
        BloodResultsLipidsAction,
        BloodResultsRftAction,
    ]:
        try:
            django_apps.get_model(action_cls.reference_model)
        except LookupError:
            pass
        else:
            site_action_items.register(action_cls)


register_actions()



from django import forms

from ..calculators import CalculatorError, calculate_bmi

__all__ = ["BmiFormValidatorMixin"]


class BmiFormValidatorMixin:
    def validate_bmi(self, **kwargs):
        try:
            bmi = calculate_bmi(
                height_cm=self.cleaned_data.get("height"),
                weight_kg=self.cleaned_data.get("weight"),
                **self.cleaned_data,
                **kwargs,
            )
        except CalculatorError as e:
            raise forms.ValidationError(str(e)) from e
        return bmi



from django.db import models

from ..calculators import calculate_bmi
from ..models import HeightField, WeightField


class WeightHeightBmiModelMixin(models.Model):
    lower_bmi_value = 5.0

    upper_bmi_value = 60.0

    weight = WeightField(null=True, blank=True)

    height = HeightField(null=True, blank=True)

    calculated_bmi_value = models.DecimalField(
        verbose_name="BMI",
        max_digits=8,
        decimal_places=4,
        null=True,
        blank=False,
        help_text="system calculated",
    )

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        bmi = calculate_bmi(
            weight_kg=self.weight,
            height_cm=self.height,
            lower_bmi_value=self.lower_bmi_value,
            upper_bmi_value=self.upper_bmi_value,
            report_datetime=self.report_datetime,
            dob=self.get_dob(),
        )
        self.calculated_bmi_value = bmi.value if bmi else None
        super().save(*args, **kwargs)

    def get_dob(self):
        """Override to provides DoB if not on the model"""
        return



subject_contacts_fieldset = (
    "Subject Contact Information",
    {
        "fields": (
            "may_call",
            "may_visit_home",
            "may_sms",
            "mail_address",
            "physical_address",
            "subject_cell",
            "subject_cell_alt",
            "subject_phone",
            "subject_phone_alt",
        )
    },
)


work_contacts_fieldset = (
    "Subject's Work Contacts",
    {
        "fields": (
            "may_call_work",
            "subject_work_place",
            "subject_work_phone",
            "subject_work_cell",
        )
    },
)

indirect_contacts_fieldset = (
    "Other Contact Information",
    {
        "fields": (
            "may_contact_indirectly",
            "indirect_contact_name",
            "indirect_contact_relation",
            "indirect_contact_physical_address",
            "indirect_contact_cell",
            "indirect_contact_cell_alt",
            "indirect_contact_phone",
        )
    },
)



from typing import Any

from django.apps import apps as django_apps

from edc_constants.constants import OPEN
from edc_lab.constants import SHIPPED
from edc_lab.models import Manifest

from ..listboard_filters import PackListboardViewFilters
from .base_listboard_view import BaseListboardView

app_config = django_apps.get_app_config("edc_lab_dashboard")
edc_lab_app_config = django_apps.get_app_config("edc_lab")


class PackListboardView(BaseListboardView):
    form_action_url = "pack_form_action_url"
    listboard_url = "pack_listboard_url"
    listboard_template = "pack_listboard_template"
    listboard_model = "edc_lab.box"
    listboard_view_permission_codename = "edc_lab_dashboard.view_lab_pack_listboard"
    listboard_view_only_my_permission_codename = None
    navbar_selected_item = "pack"
    listboard_view_filters = PackListboardViewFilters()
    search_form_url = "pack_listboard_url"

    @property
    def open_manifests(self):
        return Manifest.objects.filter(status=OPEN).order_by("-manifest_datetime")

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        kwargs.update(
            new_box=self.listboard_model_cls(),
            open_manifests=self.open_manifests,
            SHIPPED=SHIPPED,
        )
        return super().get_context_data(**kwargs)



from .crf_requisition_form_validator_mixin import CrfRequisitionFormValidatorMixin
from .requisition_form_validator import RequisitionFormValidator
from .requisition_form_validator_mixin import RequisitionFormValidatorError



from __future__ import annotations

from edc_constants.constants import FEMALE, MALE
from edc_reportable import MICROMOLES_PER_LITER, MILLIGRAMS_PER_DECILITER
from edc_reportable.utils import convert_units


class EgfrCalculatorError(Exception):
    pass


class BaseEgfr:
    def __init__(
        self,
        gender: str | None = None,
        age_in_years: int | float | None = None,
        creatinine_value: float | int | None = None,
        creatinine_units: str | None = None,
        **kwargs,  # noqa
    ):
        """Expects creatinine (scr) in umols/L.

        Converts to creatinine to mg/dL for the calculation.
        """
        self.scr = {}
        if not gender or gender not in [MALE, FEMALE]:
            raise EgfrCalculatorError(
                f"Invalid gender. Expected one of {MALE}, {FEMALE}. Got {gender}."
            )
        self.gender = gender
        if not (18 <= (age_in_years or 0) < 120):
            raise EgfrCalculatorError(
                f"Invalid age. See {self.__class__.__name__}. Got {age_in_years}"
            )
        self.age_in_years = float(age_in_years) if age_in_years else None
        if creatinine_value and creatinine_units:
            self.scr.update(
                {
                    MILLIGRAMS_PER_DECILITER: convert_units(
                        label="creatinine",
                        value=float(creatinine_value),
                        units_from=creatinine_units,
                        units_to=MILLIGRAMS_PER_DECILITER,
                    )
                }
            )
            self.scr.update(
                {
                    MICROMOLES_PER_LITER: convert_units(
                        label="creatinine",
                        value=float(creatinine_value),
                        units_from=creatinine_units,
                        units_to=MICROMOLES_PER_LITER,
                    )
                }
            )



from edc_visit_schedule.models import VisitSchedule


class QueryVisitSchedule(VisitSchedule):
    def __str__(self):
        return f"{self.visit_code}: {self.visit_title}"

    class Meta:
        proxy = True
        default_permissions = ("view", "export")
        verbose_name = "Visit Schedule"
        verbose_name_plural = "Visit Schedule"



from __future__ import annotations

from edc_constants.constants import BLACK, FEMALE
from edc_reportable import MILLIGRAMS_PER_DECILITER

from .base_egrfr import BaseEgfr, EgfrCalculatorError

# TODO: https://www.rcpa.edu.au/Manuals/RCPA-Manual/
#  Pathology-Tests/C/Creatinine-clearance-Cockcroft-and-Gault


class EgfrCkdEpi(BaseEgfr):
    """Reference https://nephron.com/epi_equation

    CKD-EPI Creatinine equation (2009)

    Levey AS, Stevens LA, et al. A New Equation to Estimate Glomerular
    Filtration Rate. Ann Intern Med. 2009; 150:604-612.
    """

    def __init__(self, ethnicity: str = None, **kwargs):
        self.ethnicity = ethnicity
        super().__init__(**kwargs)

    @property
    def value(self) -> float | None:
        if (
            self.gender
            and self.age_in_years
            and self.ethnicity
            and self.scr.get(MILLIGRAMS_PER_DECILITER)
        ):
            scr = self.scr.get(MILLIGRAMS_PER_DECILITER)
            return float(
                141.000
                * (min(scr / self.kappa, 1.000) ** self.alpha)
                * (max(scr / self.kappa, 1.000) ** -1.209)
                * self.age_factor
                * self.gender_factor
                * self.ethnicity_factor
            )
        opts = dict(
            gender=self.gender,
            age_in_years=self.age_in_years,
            ethnicity=self.ethnicity,
            scr=self.scr.get(MILLIGRAMS_PER_DECILITER),
        )
        raise EgfrCalculatorError(f"Unable to calculate. Insufficient information. Got {opts}")

    @property
    def alpha(self) -> float:
        return float(-0.329 if self.gender == FEMALE else -0.411)

    @property
    def kappa(self) -> float:
        return float(0.7 if self.gender == FEMALE else 0.9)

    @property
    def ethnicity_factor(self) -> float:
        return float(1.159 if self.ethnicity == BLACK else 1.000)

    @property
    def gender_factor(self) -> float:
        return float(1.018 if self.gender == FEMALE else 1.000)

    @property
    def age_factor(self) -> float:
        return float(0.993**self.age_in_years)



from django.db import models


class LocatorMethodsModelMixin(models.Model):
    @property
    def call(self):
        return self.may_call

    @property
    def call_work(self):
        return self.may_call_work

    @property
    def visit_home(self):
        return self.may_visit_home

    @property
    def sms(self):
        return self.may_sms

    @property
    def contact_indirectly(self):
        return self.may_contact_indirectly

    class Meta:
        abstract = True



class RsDialect:
    def __init__(self, obj=None):
        self.obj = obj

    @property
    def select_registered_subject(self):
        """Returns an SQL statement that joins visit, appt, and registered_subject."""
        sql = (  # nosec B608
            "SELECT R.subject_identifier, R.screening_identifier, R.dob, "  # nosec B608
            "R.gender, R.subject_type, R.sid, "  # nosec B608
            "R.screening_age_in_years, R.registration_status, "  # nosec B608
            "R.registration_datetime, "  # nosec B608
            "R.randomization_datetime "  # nosec B608
            f"FROM {self.obj.registered_subject_tbl} as R "  # nosec B608
        )
        return sql, None



from typing import Any

from django.views.generic.base import ContextMixin

from .research_protocol_config import ResearchProtocolConfig


class EdcProtocolViewMixin(ContextMixin):
    def get_context_data(self, **kwargs) -> dict[str, Any]:
        protocol_config = ResearchProtocolConfig()
        kwargs.update(
            {
                "protocol": protocol_config.protocol,
                "protocol_number": protocol_config.protocol_number,
                "protocol_name": protocol_config.protocol_name,
                "protocol_title": protocol_config.protocol_title,
            }
        )
        return super().get_context_data(**kwargs)



from django import forms
from django.apps import apps as django_apps
from django.core.exceptions import ImproperlyConfigured

from .constants import CLOSED_TIMEPOINT


class TimepointFormMixin:
    def clean(self):
        cleaned_data = super().clean()
        app_config = django_apps.get_app_config("edc_timepoint")
        try:
            app_config.timepoints[self._meta.model._meta.label_lower]
        except KeyError as e:
            raise ImproperlyConfigured(
                "ModelForm uses a model that is not a timepoint. "
                f"Got {self._meta.model._meta.label_lower}."
            ) from e
        timepoint_status = cleaned_data.get("timepoint_status")
        if timepoint_status == CLOSED_TIMEPOINT:
            raise forms.ValidationError(
                f"This '{self._meta.verbose_name}' record is closed "
                "for data entry. See Timepoint."
            )
        return cleaned_data



from django.db import models
from django.db.models import PROTECT
from django.utils.translation import gettext_lazy as _

from edc_appointment.utils import get_appointment_model_name
from edc_consent.model_mixins import RequiresConsentFieldsModelMixin
from edc_metadata.model_mixins.creates import CreatesMetadataModelMixin
from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_sites.managers import CurrentSiteManager
from edc_sites.model_mixins import SiteModelMixin
from edc_visit_tracking.choices import (
    VISIT_INFO_SOURCE,
    VISIT_REASON,
    VISIT_REASON_MISSED,
)
from edc_visit_tracking.managers import VisitModelManager
from edc_visit_tracking.model_mixins import VisitModelMixin


class SubjectVisit(
    VisitModelMixin,
    CreatesMetadataModelMixin,
    SiteModelMixin,
    RequiresConsentFieldsModelMixin,
    BaseUuidModel,
):
    appointment = models.OneToOneField(
        get_appointment_model_name(),
        on_delete=PROTECT,
        related_name="default_subjectvisit",
    )

    reason = models.CharField(max_length=25, choices=VISIT_REASON)

    reason_missed = models.CharField(
        verbose_name=_("If 'missed', provide the reason for the missed visit"),
        max_length=35,
        choices=VISIT_REASON_MISSED,
        blank=True,
        default="",
    )

    info_source = models.CharField(
        verbose_name=_("What is the main source of this information?"),
        max_length=25,
        choices=VISIT_INFO_SOURCE,
    )

    objects = VisitModelManager()

    on_site = CurrentSiteManager()

    history = HistoricalRecords()

    class Meta(VisitModelMixin.Meta, BaseUuidModel.Meta):
        indexes = (*VisitModelMixin.Meta.indexes, *BaseUuidModel.Meta.indexes)



import pandas as pd

from ..constants import SYSTEM_COLUMNS
from ..df_handlers import NonCrfDfHandler
from .csv_crf_tables_exporter import CsvCrfTablesExporter


class CsvCrfInlineTablesExporter(CsvCrfTablesExporter):
    """A class to export non-CRF tables for this app_label."""

    visit_column = None
    df_inline_handler_cls = NonCrfDfHandler
    exclude_inline_tables = None

    def __init__(self, exclude_inline_tables=None, **kwargs):
        self.exclude_inline_tables = exclude_inline_tables or self.exclude_inline_tables or []
        super().__init__(**kwargs)

    def to_csv(self, export_folder=None, **kwargs):
        """Exports all tables to CSV."""
        self.exported_paths = {}
        export_folder = export_folder or self.export_folder
        for table_name in self.table_names:
            df_table = self.to_df(table_name=table_name, exclude_system_columns=True, **kwargs)
            for row in self.get_inline_table_names(table_name).itertuples():
                if row.table_name not in self.exclude_inline_tables:
                    df_inline = self.to_inline_df(table_name=row.table_name, **kwargs)
                    df = self.merge_with_inline(
                        df_table,
                        df_inline,
                        left_on=row.referenced_column_name,
                        right_on=row.column_name,
                    )
                    df = self.get_prepped_inline_df(table_name=row.table_name, df=df)
                    label = row.column_name.replace("_id", "").replace("_", "")
                    csv_exporter = self.csv_exporter_cls(
                        data_label=f"{row.table_name}_{label}_merged",
                        export_folder=export_folder,
                        **kwargs,
                    )
                    exported = csv_exporter.to_csv(dataframe=df)
                    if exported.path:
                        self.exported_paths.update({table_name: exported.path})

    def get_inline_table_names(self, table_name=None):
        return self.db.show_inline_tables(referenced_table_name=table_name)

    def merge_with_inline(self, df_table=None, df_inline=None, left_on=None, right_on=None):
        """Merge main table and its inline."""
        df_table = df_table[[col for col in df_table.columns if col not in SYSTEM_COLUMNS]]
        df = pd.merge(
            df_table,
            df_inline,
            how="left",
            left_on=left_on,
            right_on=right_on,
            suffixes=["_master", ""],
        )
        return df

    def to_inline_df(self, table_name=None, **kwargs):
        """Returns a "prepped" dataframe for this inline table_name."""
        df = self.get_raw_df(table_name)
        return self.get_prepped_inline_df(table_name, df, **kwargs)

    def get_prepped_inline_df(self, table_name=None, df=None, **kwargs):
        """Returns a dataframe after passing the given inline df
        through the inline df_handler class.
        """
        if self.df_handler_cls:
            df_handler = self.df_inline_handler_cls(
                dataframe=df, db=self.db, table_name=table_name, **kwargs
            )
            df = df_handler.dataframe
        return df



from django.apps import apps as django_apps

from edc_list_data.model_mixins import ListModelMixin


class ModelOptions(dict):
    """A serializable object of a selection of model attrs."""

    def __init__(self, model: str = None):
        model_cls = django_apps.get_model(model)
        self.model: str = str(model)
        self.app_label: str = str(model_cls._meta.app_label)
        self.app_name: str = " ".join(model_cls._meta.app_label.split("_")).title()
        self.verbose_name: str = str(model_cls._meta.verbose_name)
        self.label_lower: str = str(model_cls._meta.label_lower)
        self.fields: list[str] = [f.name for f in model_cls._meta.get_fields()]
        self.is_historical: bool = model_cls._meta.label_lower.split(".")[1].startswith(
            "historical"
        )
        self.is_list_model: bool = issubclass(model_cls, (ListModelMixin,))
        self.db_table: str = str(model_cls._meta.db_table)
        dict.__init__(
            self,
            app_label=self.app_label,
            app_name=self.app_name,
            model=self.model,
            verbose_name=self.verbose_name,
            label_lower=self.label_lower,
            is_historical=self.is_historical,
            is_list_model=self.is_list_model,
            is_inline=self.is_inline,
            db_table=self.db_table,
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(model='{self.label_lower}')"

    @property
    def is_inline(self) -> bool:
        return False



from __future__ import annotations

from typing import TYPE_CHECKING

from ...view_mixins import BoxViewMixin
from .base_listboard_view import BaseListboardView

if TYPE_CHECKING:
    from django.db.models import Q


class BaseBoxItemListboardView(BoxViewMixin, BaseListboardView):
    navbar_selected_item = "pack"
    ordering = ("-position",)
    listboard_model = "edc_lab.boxitem"
    listboard_view_permission_codename = "edc_lab_dashboard.view_lab_box_listboard"

    def get_queryset_filter_options(self, request, *args, **kwargs) -> tuple[Q, dict]:
        q_object, options = super().get_queryset_filter_options(request, *args, **kwargs)
        options.update({"box": self.box})
        return q_object, {"box": self.box}



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING
from uuid import UUID

from .model_button import ModelButton

if TYPE_CHECKING:
    from django.contrib.sites.models import Site

    from edc_appointment.models import Appointment
    from edc_crf.model_mixins import CrfModelMixin
    from edc_lab.model_mixins import RequisitionModelMixin
    from edc_metadata.models import CrfMetadata, RequisitionMetadata
    from edc_model.models import BaseUuidModel

    class CrfModel(CrfModelMixin, BaseUuidModel): ...

    class RequisitionModel(RequisitionModelMixin, BaseUuidModel): ...


__all__ = ["DashboardModelButton"]


@dataclass
class DashboardModelButton(ModelButton):
    """
    Base class for buttons that appear on the subject dashboard.

    Note:
        dashboard_url_name: key in global `url_names`
        See edc_dashboard.

    """

    model_obj: CrfModel | RequisitionModel = field(init=False)
    metadata_model_obj: CrfMetadata | RequisitionMetadata = None
    appointment: Appointment = None
    next_url_name: str = field(default="subject_dashboard_url")
    model_cls: type[CrfModel | RequisitionModel] = field(default=None, init=False)

    def __post_init__(self):
        self.model_cls = self.metadata_model_obj.model_cls
        self.model_obj = self.metadata_model_obj.model_instance

    @property
    def site(self) -> Site | None:
        """If model_obj is None, then Site should come from the
        CRFMetadata or RequisitionMetadata models.
        """
        return getattr(self.model_obj, "site", None) or getattr(
            self.metadata_model_obj, "site", None
        )

    @property
    def reverse_kwargs(self) -> dict[str, str | UUID]:
        return dict(
            subject_identifier=self.appointment.subject_identifier,
            appointment=self.appointment.id,
        )



# Generated by Django 4.2.7 on 2023-12-03 00:22

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_facility", "0008_alter_healthfacility_device_created_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="healthfacility",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Health Facility",
                "verbose_name_plural": "Health Facilities",
            },
        ),
        migrations.AlterModelOptions(
            name="healthfacilitytypes",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Health Facility Type",
                "verbose_name_plural": "Health Facility Types",
            },
        ),
        migrations.AlterModelOptions(
            name="holiday",
            options={"verbose_name": "Holiday", "verbose_name_plural": "Holidays"},
        ),
        migrations.RemoveIndex(
            model_name="healthfacilitytypes",
            name="edc_facilit_id_05cd80_idx",
        ),
        migrations.AddField(
            model_name="healthfacility",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="healthfacility",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalhealthfacility",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalhealthfacility",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="healthfacilitytypes",
            name="display_index",
            field=models.IntegerField(
                default=0,
                help_text="Index to control display order if not alphabetical, not required",
                verbose_name="display index",
            ),
        ),
        migrations.AlterField(
            model_name="healthfacilitytypes",
            name="display_name",
            field=models.CharField(
                help_text="(suggest 40 characters max.)",
                max_length=250,
                unique=True,
                verbose_name="Name",
            ),
        ),
        migrations.AlterField(
            model_name="healthfacilitytypes",
            name="name",
            field=models.CharField(
                help_text="This is the stored value, required",
                max_length=250,
                unique=True,
                verbose_name="Stored value",
            ),
        ),
        migrations.AddIndex(
            model_name="healthfacility",
            index=models.Index(
                fields=["-modified", "-created"], name="edc_facilit_modifie_dcf8ac_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="healthfacility",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_facilit_user_mo_c7241a_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="healthfacilitytypes",
            index=models.Index(fields=["name"], name="edc_facilit_name_78007a_idx"),
        ),
        migrations.AddIndex(
            model_name="healthfacilitytypes",
            index=models.Index(
                fields=["display_index", "display_name"],
                name="edc_facilit_display_56fc12_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="holiday",
            index=models.Index(
                fields=["name", "country", "local_date"],
                name="edc_facilit_name_acd5e1_idx",
            ),
        ),
    ]



ALREADY_REPORTED = "Already reported"
INVALID_REFERENCE = "invalid_reference"
MILD = "mild"
MODERATE = "moderate"
PRESENT_AT_BASELINE = "present_at_baseline"
SEVERE = "severe"
SEVERITY_INCREASED_FROM_G3 = "increase_from_g3"
LLN = "LLN"
ULN = "ULN"
HIGH_VALUE = "9999999999.9"



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from edc_action_item.fieldsets import action_fieldset_tuple


class SubjectTransferModelAdminMixin:
    form = None

    fieldsets = (
        (None, {"fields": ("subject_identifier", "report_datetime")}),
        (
            "Transfer Details",
            {
                "fields": (
                    "transfer_date",
                    "initiated_by",
                    "initiated_by_other",
                    "transfer_reason",
                    "transfer_reason_other",
                    "may_return",
                    "may_contact",
                    "comment",
                )
            },
        ),
        action_fieldset_tuple,
        audit_fieldset_tuple,
    )

    filter_horizontal = ("transfer_reason",)

    radio_fields = {  # noqa: RUF012
        "initiated_by": admin.VERTICAL,
        "may_return": admin.VERTICAL,
        "may_contact": admin.VERTICAL,
    }

    def get_list_display(self, request) -> tuple[str, ...]:
        list_display = super().get_list_display(request)
        custom_fields = (
            "subject_identifier",
            "dashboard",
            "transfer_date",
            "initiated_by",
            "may_return",
            "may_contact",
        )
        return custom_fields + tuple(f for f in list_display if f not in custom_fields)

    def get_list_filter(self, request) -> tuple[str, ...]:
        list_filter = super().get_list_filter(request)
        custom_fields = (
            "transfer_date",
            "initiated_by",
            "may_return",
            "may_contact",
        )
        return custom_fields + tuple(f for f in list_filter if f not in custom_fields)

    def get_search_fields(self, request) -> tuple[str, ...]:
        search_fields = super().get_search_fields(request)
        custom_fields = ("subject_identifier",)
        return tuple(set(custom_fields + search_fields))



from __future__ import annotations

import copy
import sys
from importlib import import_module
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.conf import settings
from django.core.management.color import color_style
from django.urls import NoReverseMatch
from django.utils.module_loading import module_has_submodule

from .exceptions import AlreadyRegistered, NavbarError

if TYPE_CHECKING:
    from .navbar import Navbar


class NavbarCollection:
    """A class to contain a dictionary of navbars. See Navbar."""

    name = "default"

    def __init__(self):
        self.registry = {}
        self.codenames = {}

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def register(self, navbar: Navbar = None):
        if navbar.name not in self.registry:
            self.registry.update({navbar.name: navbar})
            # self.codenames.update(**navbar.codenames)
        else:
            raise AlreadyRegistered(f"Navbar with name {navbar.name} is already registered.")

    def context(self, name=None, selected_item=None):
        """Returns the named navbar in the collection as context."""
        return dict(
            navbar_item_selected=selected_item,
            navbar=self.get_navbar(name=name, selected_item=selected_item),
            navbar_name=name,
        )

    def get_navbar(self, name: str = None, selected_item: str = None) -> Navbar:
        """Returns a selected navbar in the collection."""
        # does the navbar exist?
        try:
            navbar: Navbar = self.registry[name]
        except KeyError:
            raise NavbarError(
                f"Navbar '{name}' does not exist. Expected one of "
                f"{list(self.registry.keys())}. See {self!r}."
            )
        else:
            # does the navbar have items?
            if not navbar.navbar_items:
                raise NavbarError(
                    f"Navbar '{navbar.name}' has no navbar_item. Expected "
                    f"'{selected_item}'. See {self!r}"
                )
            # does the selected item exist?
            if selected_item:
                navbar.set_active(selected_item)
        return navbar

    def show_user_permissions(self, username: str = None, navbar_name: str = None):
        user = django_apps.get_model("auth.user").objects.get(username=username)
        navbar: Navbar = self.registry.get(navbar_name)
        return navbar.show_user_permissions(user=user)

    def show_user_codenames(self, username=None, navbar_name=None):
        user_permissions = self.show_user_permissions(username, navbar_name)
        codenames = []
        for codename in [list(v.keys()) for v in user_permissions.values()]:
            codenames.extend(codename)
        return codenames

    def autodiscover(self, module_name=None, verbose=True):
        if getattr(settings, "EDC_NAVBAR_ENABLED", True):
            module_name = module_name or "navbars"
            writer = sys.stdout.write if verbose else lambda x: x
            style = color_style()
            writer(f" * checking for site {module_name} ...\n")
            for app in django_apps.app_configs:
                try:
                    mod = import_module(app)
                    try:
                        before_import_registry = copy.copy(site_navbars.registry)
                        import_module(f"{app}.{module_name}")
                        writer(f"   - registered navbars '{module_name}' from '{app}'\n")
                    except NavbarError as e:
                        writer(f"   * loading {app}.navbars ... \n")
                        writer(style.ERROR(f"ERROR! {e}\n"))
                    except NoReverseMatch as e:
                        writer(f"   * loading {app}.navbars ... \n")
                        writer(style.ERROR(f"ERROR! {e}\n"))
                        raise
                    except ImportError as e:
                        site_navbars.registry = before_import_registry
                        if module_has_submodule(mod, module_name):
                            raise NavbarError(e)
                except ImportError:
                    pass


site_navbars = NavbarCollection()



from .get_appointments import get_appointments
from .get_crf import get_crf
from .get_eos import get_eos
from .get_next_appointments import get_next_appointments
from .get_subject_consent import get_subject_consent
from .get_subject_visit import get_subject_visit



from __future__ import annotations

from typing import TYPE_CHECKING

from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_utils.text import convert_from_camel

from ..url_config import UrlConfig
from ..url_names import InvalidDashboardUrlName, url_names

if TYPE_CHECKING:
    from django.urls import URLPattern


class UrlRequestContextError(Exception):
    pass


class UrlRequestContextMixin:
    urlconfig_getattr = "dashboard_urls"
    urlconfig_identifier_label = "subject_identifier"
    urlconfig_identifier_pattern = ResearchProtocolConfig().subject_identifier_pattern
    urlconfig_label = None
    url_name = None

    @classmethod
    def get_urlname(cls):
        return cls.url_name

    @classmethod
    def urls(
        cls,
        *,
        label: str,
        identifier_pattern: str,
        namespace: str | None = None,
        identifier_label: str | None = None,
    ) -> list[URLPattern]:
        label = (
            label
            or cls.urlconfig_label
            or convert_from_camel(cls.__name__.replace("view", "")).lower()
        )
        urlconfig = UrlConfig(
            url_name=cls.get_urlname(),
            namespace=namespace,
            view_class=cls,
            label=label,
            identifier_label=identifier_label or cls.urlconfig_identifier_label,
            identifier_pattern=identifier_pattern or cls.urlconfig_identifier_pattern,
        )
        return getattr(urlconfig, cls.urlconfig_getattr)

    @staticmethod
    def add_url_to_context(new_key=None, existing_key=None) -> dict[str, str]:
        """Add url as new_key to the context using the value
        of the existing_key from request.context_data.
        """
        try:
            url_data = {new_key: url_names.get(existing_key)}
        except InvalidDashboardUrlName as e:
            raise UrlRequestContextError(
                f"Url name not defined in url_names. "
                f"Expected one of {url_names.registry}. Got {e}. "
                f"Hint: check if dashboard middleware is loaded."
            ) from e
        return url_data



from __future__ import annotations

from datetime import date, datetime

from django.utils import timezone

from edc_utils import age

from .evaluator import Evaluator


class AgeEvaluator(Evaluator):
    def __init__(
        self,
        age_lower: int | None = None,
        age_upper: int = None,
        age_units: str | None = None,
        age_lower_inclusive: bool | None = None,
        age_upper_inclusive: bool | None = None,
        **kwargs,
    ):
        kwargs["units"] = age_units or "years"
        kwargs["lower"] = age_lower or 0
        kwargs["upper"] = age_upper
        kwargs["lower_inclusive"] = age_lower_inclusive
        kwargs["upper_inclusive"] = age_upper_inclusive
        super().__init__(**kwargs)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.description(show_as_int=True)})"

    def description(self, value: int = None, **kwargs):
        kwargs["show_as_int"] = True
        kwargs["placeholder"] = "AGE"
        return super().description(value=value, **kwargs)

    def in_bounds_or_raise(
        self,
        dob: date = None,
        report_datetime: datetime = None,
        age_units: str | None = None,
    ) -> bool:
        report_datetime = report_datetime or timezone.now()
        age_units = age_units or "years"
        rdelta = age(dob, report_datetime)
        value = getattr(rdelta, self.units)
        return super().in_bounds_or_raise(value, units=age_units)



from __future__ import annotations

from typing import TYPE_CHECKING
from zoneinfo import ZoneInfo

from django.conf import settings
from django.db import models
from django.utils import timezone

from edc_identifier.managers import SubjectIdentifierManager
from edc_identifier.model_mixins import UniqueSubjectIdentifierFieldMixin
from edc_model.models import HistoricalRecords
from edc_model.validators import datetime_not_future
from edc_protocol.validators import datetime_not_before_study_start
from edc_utils import convert_php_dateformat

from ..site_visit_schedules import site_visit_schedules

if TYPE_CHECKING:
    from datetime import datetime

    from ..schedule import Schedule
    from ..visit_schedule import VisitSchedule


class OnScheduleManager(SubjectIdentifierManager):
    def put_on_schedule(
        self,
        subject_identifier: str,
        onschedule_datetime: datetime,
        skip_get_current_site: bool | None = None,
    ) -> None:
        """Puts a subject on the schedule associated with this
        OnSchedule model.

        Note: So far this is mostly used in tests. It is just as easy
        to fetch the schedule by name and call
        `schedule.put_on_schedule`
        """
        _, schedule = site_visit_schedules.get_by_onschedule_model(
            self.model._meta.label_lower
        )
        schedule.put_on_schedule(
            subject_identifier=subject_identifier,
            onschedule_datetime=onschedule_datetime,
            skip_get_current_site=skip_get_current_site,
        )


class OnScheduleModelMixin(UniqueSubjectIdentifierFieldMixin, models.Model):
    """A model mixin for a schedule's onschedule model."""

    onschedule_datetime = models.DateTimeField(
        validators=[datetime_not_before_study_start, datetime_not_future],
        default=timezone.now,
    )

    report_datetime = models.DateTimeField(editable=False)

    objects = OnScheduleManager()

    history = HistoricalRecords(inherit=True)

    def __str__(self):
        formatted_datetime = self.report_datetime.astimezone(
            ZoneInfo(settings.TIME_ZONE)
        ).strftime(convert_php_dateformat(settings.SHORT_DATETIME_FORMAT))
        return f"{self.subject_identifier} {formatted_datetime}"

    def natural_key(self):
        return (self.subject_identifier,)

    def save(self, *args, **kwargs):
        self.report_datetime = self.onschedule_datetime
        super().save(*args, **kwargs)

    @property
    def visit_schedule(self) -> VisitSchedule:
        """Returns a visit schedule object."""
        return site_visit_schedules.get_by_onschedule_model(
            onschedule_model=self._meta.label_lower
        )[0]

    @property
    def schedule(self) -> Schedule:
        """Returns a schedule object."""
        return site_visit_schedules.get_by_onschedule_model(
            onschedule_model=self._meta.label_lower
        )[1]

    class Meta:
        abstract = True
        indexes = (models.Index(fields=["subject_identifier", "onschedule_datetime", "site"]),)



from __future__ import annotations

from typing import TYPE_CHECKING

from django.contrib import admin
from django.core.exceptions import ObjectDoesNotExist
from django_audit_fields.admin import audit_fieldset_tuple

from edc_appointment.utils import get_appointment_model_cls
from edc_constants.constants import OTHER
from edc_document_status.fieldsets import document_status_fieldset_tuple
from edc_document_status.modeladmin_mixins import DocumentStatusModelAdminMixin
from edc_visit_schedule.fieldsets import (
    visit_schedule_fields,
    visit_schedule_fieldset_tuple,
)

from ..constants import SCHEDULED, UNSCHEDULED

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


class VisitModelAdminMixin(DocumentStatusModelAdminMixin):
    """ModelAdmin subclass for models with a ForeignKey to
    'appointment', such as your visit model(s).

    In the child ModelAdmin class set the following attributes,
    for example:

        visit_attr = 'maternal_visit'
        dashboard_type = 'maternal'
    """

    date_hierarchy = "report_datetime"

    fieldsets = (
        (
            None,
            {
                "fields": [
                    "appointment",
                    "report_datetime",
                    "reason",
                    "reason_missed",
                    "reason_unscheduled",
                    "reason_unscheduled_other",
                    "info_source",
                    "info_source_other",
                    "comments",
                ]
            },
        ),
        visit_schedule_fieldset_tuple,
        document_status_fieldset_tuple,
        audit_fieldset_tuple,
    )

    radio_fields = {  # noqa: RUF012
        "reason": admin.VERTICAL,
        "reason_unscheduled": admin.VERTICAL,
        "reason_missed": admin.VERTICAL,
        "info_source": admin.VERTICAL,
        "require_crfs": admin.VERTICAL,
    }

    search_fields = (
        "id",
        "reason",
        "visit_code",
        "subject_identifier",
    )

    @staticmethod
    def subject_identifier(obj=None) -> str:
        return obj.appointment.subject_identifier

    @staticmethod
    def visit_reason(obj=None) -> str:
        if obj.reason != UNSCHEDULED:
            visit_reason = obj.get_reason_display()
        elif obj.reason_unscheduled == OTHER:
            visit_reason = obj.reason_unscheduled_other
        else:
            visit_reason = obj.get_reason_unscheduled_display()
        return visit_reason

    @staticmethod
    def status(obj=None) -> str:
        return obj.study_status

    @staticmethod
    def scheduled_data(obj=None) -> str:
        return obj.get_require_crfs_display()

    def formfield_for_foreignkey(self, db_field, request: WSGIRequest, **kwargs):
        db = kwargs.get("using")
        if db_field.name == "appointment" and request.GET.get("appointment"):
            kwargs["queryset"] = db_field.related_model._default_manager.using(db).filter(
                pk=request.GET.get("appointment")
            )
        else:
            kwargs["queryset"] = db_field.related_model._default_manager.none()
        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    def get_list_display(self, request: WSGIRequest) -> tuple[str, ...]:
        list_display = super().get_list_display(request)
        custom_fields = (
            "appointment",
            "subject_identifier",
            "report_datetime",
            "visit_reason",
            "status",
            "scheduled_data",
        )
        return *custom_fields, *tuple(f for f in list_display if f not in custom_fields)

    def get_list_filter(self, request: WSGIRequest) -> tuple[str, ...]:
        list_filter = super().get_list_filter(request)
        custom_fields = (
            "report_datetime",
            "visit_code",
            "visit_code_sequence",
            "reason",
            "require_crfs",
        )
        return *custom_fields, *tuple(f for f in list_filter if f not in custom_fields)

    def get_readonly_fields(self, request, obj=None) -> tuple[str, ...]:
        readonly_fields = super().get_readonly_fields(request, obj=obj)
        return tuple(set(readonly_fields + visit_schedule_fields))

    def get_changeform_initial_data(self, request: WSGIRequest) -> dict:
        """Sets initial data for the form.

        Inherit from this add additional fields to set.

        Gets report_datetime from the appointment.appt_datetime
        and reason from the appointment.visit_code_sequence.
        """
        initial_data = super().get_changeform_initial_data(request)
        appointment_id = request.GET.get("appointment")
        try:
            appointment = get_appointment_model_cls().objects.get(id=appointment_id)
        except ObjectDoesNotExist:
            initial_data.update(
                report_datetime=None,
                reason=None,
            )
        else:
            initial_data.update(
                report_datetime=appointment.appt_datetime,
                reason=(SCHEDULED if appointment.visit_code_sequence == 0 else UNSCHEDULED),
            )
        return initial_data



from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models

from edc_lab_panel.constants import (
    HAEMOGLOBIN,
    HCT,
    LYMPHOCYTE,
    LYMPHOCYTE_DIFF,
    MCH,
    MCHC,
    MCV,
    NEUTROPHIL,
    NEUTROPHIL_DIFF,
    PLATELETS,
    RBC,
    WBC,
)
from edc_reportable import GRAMS_PER_DECILITER, PERCENT
from edc_reportable.units import (
    CELLS_PER_MILLIMETER_CUBED,
    CELLS_PER_MILLIMETER_CUBED_DISPLAY,
    FEMTOLITERS_PER_CELL,
    PICOGRAMS_PER_CELL,
    TEN_X_9_PER_LITER,
)

from ..model_mixin_factories import reportable_result_model_mixin_factory


class HaemoglobinModelMixin(
    reportable_result_model_mixin_factory(
        utest_id=HAEMOGLOBIN,
        verbose_name="Haemoglobin",
        units_choices=((GRAMS_PER_DECILITER, GRAMS_PER_DECILITER),),
        decimal_places=1,
    ),
    models.Model,
):
    class Meta:
        abstract = True


class HctModelMixin(
    reportable_result_model_mixin_factory(
        utest_id=HCT,
        verbose_name="Hematocrit",
        units_choices=((PERCENT, PERCENT),),
        validators=[MinValueValidator(1.0), MaxValueValidator(999.0)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class MchModelMixin(
    reportable_result_model_mixin_factory(
        utest_id=MCH,
        units_choices=((PICOGRAMS_PER_CELL, PICOGRAMS_PER_CELL),),
    ),
    models.Model,
):
    class Meta:
        abstract = True


class MchcModelMixin(
    reportable_result_model_mixin_factory(
        utest_id=MCHC,
        units_choices=((GRAMS_PER_DECILITER, GRAMS_PER_DECILITER),),
    ),
    models.Model,
):
    class Meta:
        abstract = True


class McvModelMixin(
    reportable_result_model_mixin_factory(
        utest_id=MCV,
        units_choices=((FEMTOLITERS_PER_CELL, FEMTOLITERS_PER_CELL),),
    ),
    models.Model,
):
    class Meta:
        abstract = True


class NeutrophilModelMixin(
    reportable_result_model_mixin_factory(
        utest_id=NEUTROPHIL,
        verbose_name="Neutrophil (abs)",
        units_choices=((TEN_X_9_PER_LITER, TEN_X_9_PER_LITER),),
        decimal_places=2,
        validators=[MinValueValidator(0.00), MaxValueValidator(9999)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class NeutrophilDiffModelMixin(
    reportable_result_model_mixin_factory(
        utest_id=NEUTROPHIL_DIFF,
        verbose_name="Neutrophil (diff)",
        units_choices=((PERCENT, PERCENT),),
        decimal_places=2,
        validators=[MinValueValidator(0.00), MaxValueValidator(9999)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class LymphocyteModelMixin(
    reportable_result_model_mixin_factory(
        utest_id=LYMPHOCYTE,
        verbose_name="Lymphocyte (abs)",
        units_choices=((TEN_X_9_PER_LITER, TEN_X_9_PER_LITER),),
        decimal_places=2,
        validators=[MinValueValidator(0.00), MaxValueValidator(9999)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class LymphocyteDiffModelMixin(
    reportable_result_model_mixin_factory(
        utest_id=LYMPHOCYTE_DIFF,
        verbose_name="Lymphocyte (diff)",
        units_choices=((PERCENT, PERCENT),),
        decimal_places=2,
        validators=[MinValueValidator(0.00), MaxValueValidator(9999)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class PlateletsModelMixin(
    reportable_result_model_mixin_factory(
        utest_id=PLATELETS,
        verbose_name="Platelets",
        units_choices=(
            (TEN_X_9_PER_LITER, TEN_X_9_PER_LITER),
            (CELLS_PER_MILLIMETER_CUBED, CELLS_PER_MILLIMETER_CUBED_DISPLAY),
        ),
        decimal_places=0,
        validators=[MinValueValidator(1), MaxValueValidator(9999)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class RbcModelMixin(
    reportable_result_model_mixin_factory(
        utest_id=RBC,
        verbose_name="Red blood cell count",
        units_choices=(
            (TEN_X_9_PER_LITER, TEN_X_9_PER_LITER),
            (CELLS_PER_MILLIMETER_CUBED, CELLS_PER_MILLIMETER_CUBED),
        ),
        validators=[MinValueValidator(1.0), MaxValueValidator(999999.0)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class WbcModelMixin(
    reportable_result_model_mixin_factory(
        utest_id=WBC,
        units_choices=(
            (TEN_X_9_PER_LITER, TEN_X_9_PER_LITER),
            (CELLS_PER_MILLIMETER_CUBED, CELLS_PER_MILLIMETER_CUBED_DISPLAY),
        ),
    ),
    models.Model,
):
    class Meta:
        abstract = True



from django import dispatch
from django.db.models import expressions
from simple_history import signals


@dispatch.receiver(signals.pre_create_historical_record, dispatch_uid="simple_history_refresh")
def remove_f_expressions(sender, instance, history_instance, **kwargs) -> None:
    """Model with history manager fails to save if update contains
    an F() expression.

    You get a ValueError:
        ... F() expressions can only be used to update, not to insert.

    solution from:
    https://stackoverflow.com/questions/62343627/circumventing-f-expression-problem-of-
    django-simple-history-by-overriding-save/62369328#62369328
    """
    f_expression_fields = []
    for field in history_instance._meta.fields:
        field_value = getattr(history_instance, field.name)
        if isinstance(field_value, expressions.BaseExpression):
            f_expression_fields.append(field.name)

    if f_expression_fields:
        instance.refresh_from_db()
        for field_name in f_expression_fields:
            field_value = getattr(instance, field_name)
            setattr(history_instance, field_name, field_value)



from .csv_crf_inline_tables_exporter import CsvCrfInlineTablesExporter
from .csv_crf_tables_exporter import CsvCrfTablesExporter
from .csv_exporter import CsvExporter, Exporter
from .csv_model_exporter import CsvModelExporter
from .csv_noncrf_tables_exporter import CsvNonCrfTablesExporter
from .tables_exporter import TablesExporter



from django.urls import path

from .admin_site import edc_data_manager_admin
from .views import HomeView

app_name = "edc_data_manager"

urlpatterns = [
    path("admin/", edc_data_manager_admin.urls),
    path("", HomeView.as_view(), name="home_url"),
]



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "edc_registration"
    verbose_name = "Edc Registration"
    app_label = "edc_registration"
    has_exportable_data = True
    include_in_administration_section = True



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING
from uuid import UUID

from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext as _

from edc_view_utils import ADD, PrnButton

if TYPE_CHECKING:
    from edc_action_item.models import ActionItem
    from edc_appointment.models import Appointment
    from edc_crf.model_mixins import CrfModelMixin
    from edc_model.models import BaseUuidModel

    class PrnModel(BaseUuidModel):
        subject_identifier: str
        ...

    class CrfModel(CrfModelMixin): ...


@dataclass
class ActionItemPopoverListItem(PrnButton):
    action_item: ActionItem = None
    category: str = None
    appointment: Appointment | None = None
    model_obj: PrnModel | CrfModel | None = field(default=None, init=False)
    model_cls: type[PrnModel | CrfModel | None] = field(default=None, init=False)

    def __post_init__(self):
        if self.category == "reference":
            self.model_obj = self.reference_obj
            self.model_cls = self.action_item.reference_model_cls
        elif self.category == "parent":
            self.model_obj = self.parent_reference_obj
            self.model_cls = self.action_item.parent_action_item.reference_model_cls
        elif self.category == "related":
            self.model_obj = self.related_reference_obj
            self.model_cls = self.action_item.related_action_item.reference_model_cls

    @property
    def reference_obj(self) -> PrnModel | CrfModel | None:
        try:
            obj = self.action_item.reference_obj
        except ObjectDoesNotExist:
            obj = None
        return obj

    @property
    def parent_reference_obj(self) -> PrnModel | CrfModel | None:
        obj = None
        if self.action_item.parent_action_item:
            try:
                obj = self.action_item.parent_action_item.reference_obj
            except ObjectDoesNotExist:
                pass
        return obj

    @property
    def related_reference_obj(self) -> PrnModel | CrfModel | None:
        obj = None
        if self.action_item.related_action_item:
            try:
                obj = self.action_item.related_action_item.reference_obj
            except ObjectDoesNotExist:
                pass
        return obj

    def label(self) -> str:
        """Fix label to the model's verbose_name."""
        if self.action == ADD:
            # prefix with word Add
            return _("Add %(verbose_name)s") % dict(
                verbose_name=self.model_cls._meta.verbose_name
            )
        return self.model_cls._meta.verbose_name

    @property
    def reverse_kwargs(self) -> dict[str, str]:
        kwargs = dict(
            subject_identifier=self.subject_identifier or self.model_obj.subject_identifier,
        )
        if self.appointment:
            # add appointment to return back to current timepoint
            # instead of schedule of all timepoints
            kwargs.update({"appointment": self.appointment.id})
        return kwargs

    @property
    def extra_kwargs(self) -> dict[str, str | int | UUID]:
        kwargs = dict(
            action_identifier=self.action_item.action_identifier,
        )
        if self.action == ADD:
            # Add field to prefill form
            for obj in [self.parent_reference_obj, self.related_reference_obj]:
                for fld in self.model_cls._meta.get_fields(include_hidden=True):
                    if fld.related_model == obj.__class__:
                        kwargs.update({fld.name: obj.id})
        return kwargs



from .site_admin import SiteAdmin
from .site_fieldset_tuple import site_fieldset_tuple
from .site_model_admin_mixin import SiteModelAdminMixin

__all__ = ["SiteAdmin", "SiteModelAdminMixin", "site_fieldset_tuple"]



from __future__ import annotations

import contextlib
import copy
import sys
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.utils.module_loading import import_module, module_has_submodule

from .exceptions import (
    AlreadyRegisteredVisitSchedule,
    RegistryNotLoaded,
    SiteVisitScheduleError,
)

if TYPE_CHECKING:
    from edc_consent.consent_definition import ConsentDefinition
    from edc_sites.single_site import SingleSite

    from .models import VisitSchedule as VisitScheduleModel
    from .schedule import Schedule
    from .visit_schedule import VisitSchedule


__all__ = ["site_visit_schedules"]


class SiteVisitSchedules:
    """Main controller of :class:`VisitSchedule` objects.

    A visit_schedule contains schedules
    """

    def __init__(self):
        self._registry: dict = {}
        self._all_post_consent_models: dict[str, str] | None = None
        self.loaded: bool = False

    @property
    def registry(self) -> dict[str, VisitSchedule]:
        if not self.loaded:
            raise RegistryNotLoaded(
                "Registry not loaded. Is AppConfig for 'edc_visit_schedule' "
                "declared in settings?."
            )
        return self._registry

    def register(self, visit_schedule: VisitSchedule) -> None:
        self.loaded = True
        if not visit_schedule.schedules:
            raise SiteVisitScheduleError(
                f"Visit schedule {visit_schedule} has no schedules. "
                f"Add one before registering."
            )
        if visit_schedule.name not in self.registry:
            self.registry.update({visit_schedule.name: visit_schedule})
        else:
            raise AlreadyRegisteredVisitSchedule(
                f"Visit Schedule {visit_schedule} is already registered."
            )
        self._all_post_consent_models = None
        self.get_offstudy_model()

    @property
    def visit_schedules(self) -> dict[str, VisitSchedule]:
        return self.registry

    def get_visit_schedule(self, visit_schedule_name=None) -> VisitSchedule:
        """Returns a visit schedule instance or raises."""
        with contextlib.suppress(AttributeError):
            visit_schedule_name = visit_schedule_name.split(".")[0]
        visit_schedule = self.registry.get(visit_schedule_name)
        if not visit_schedule:
            visit_schedule_names = "', '".join(self.registry.keys())
            raise SiteVisitScheduleError(
                f"Invalid visit schedule name. Got '{visit_schedule_name}'. "
                f"Expected one of '{visit_schedule_names}'. See {self!r}."
            )
        return visit_schedule

    def get_visit_schedules(self, *visit_schedule_names) -> dict[str, VisitSchedule]:
        """Returns a dictionary of visit schedules.

        If visit_schedule_name not specified, returns all visit schedules.
        """
        visit_schedules = {}
        for visit_schedule_name in visit_schedule_names:
            with contextlib.suppress(AttributeError):
                visit_schedule_name = visit_schedule_name.split(".")[0]  # noqa: PLW2901
            visit_schedules[visit_schedule_name] = self.get_visit_schedule(visit_schedule_name)
        return visit_schedules or self.registry

    def get_by_consent_definition(
        self, cdef: ConsentDefinition
    ) -> tuple[VisitSchedule, Schedule]:
        """Returns a visit schedule instance or raises."""
        ret = []
        attr = "consent_definitions"
        for visit_schedule in self.visit_schedules.values():
            for schedule in visit_schedule.schedules.values():
                try:
                    consent_definitions = getattr(schedule, attr)
                except (AttributeError, TypeError) as e:
                    raise SiteVisitScheduleError(
                        f"Invalid attr for Schedule. See {schedule}. Got `{attr}`."
                    ) from e
                for _cdef in consent_definitions:
                    if _cdef == cdef:
                        ret.append([visit_schedule, schedule])  # noqa: PERF401
        if not ret:
            raise SiteVisitScheduleError(
                f"Schedule not found. No schedule exists for {attr}={cdef}."
            )
        if len(ret) > 1:
            raise SiteVisitScheduleError(
                f"Schedule is ambiguous. More than one schedule exists for "
                f"{attr}={cdef}. Got {ret}"
            )
        visit_schedule, schedule = ret[0]
        return visit_schedule, schedule

    def get_by_onschedule_model(self, onschedule_model: str) -> tuple[VisitSchedule, Schedule]:
        """Returns a tuple of (visit_schedule, schedule)
        for the given onschedule model.

        attr `onschedule_model` is in "label_lower" format.
        """
        return self.get_by_model(attr="onschedule_model", model=onschedule_model)

    def get_by_offschedule_model(
        self, offschedule_model: str
    ) -> tuple[VisitSchedule, Schedule]:
        """Returns a tuple of visit_schedule, schedule
        for the given offschedule model.

        attr `offschedule_model` is in "label_lower" format.
        """
        return self.get_by_model(attr="offschedule_model", model=offschedule_model)

    def get_by_loss_to_followup_model(
        self, loss_to_followup_model: str
    ) -> tuple[VisitSchedule, Schedule]:
        """Returns a tuple of visit_schedule, schedule
        for the given loss_to_followup model.

        attr `loss_to_followup_model` is in "label_lower" format.
        """
        return self.get_by_model(attr="loss_to_followup_model", model=loss_to_followup_model)

    def get_by_model(self, attr: str, model: str) -> tuple[VisitSchedule, Schedule]:
        ret = []
        model = model.lower()
        for visit_schedule in self.visit_schedules.values():
            for schedule in visit_schedule.schedules.values():
                try:
                    model_name = getattr(schedule, attr)
                except (AttributeError, TypeError) as e:
                    raise SiteVisitScheduleError(
                        f"Invalid attr for Schedule. See {schedule}. Got {attr}."
                    ) from e
                if model_name and model_name == model:
                    ret.append([visit_schedule, schedule])
        if not ret:
            raise SiteVisitScheduleError(
                f"Schedule not found. No schedule exists for {attr}={model}."
            )
        if len(ret) > 1:
            raise SiteVisitScheduleError(
                f"Schedule is ambiguous. More than one schedule exists for "
                f"{attr}={model}. Got {ret}"
            )
        visit_schedule, schedule = ret[0]
        return visit_schedule, schedule

    def get_by_offstudy_model(self, offstudy_model: str) -> list[VisitSchedule]:
        """Returns a list of visit_schedules for the given
        offstudy model.
        """
        visit_schedules = [
            visit_schedule
            for visit_schedule in self.visit_schedules.values()
            if visit_schedule.offstudy_model == offstudy_model
        ]
        if not visit_schedules:
            raise SiteVisitScheduleError(
                f"No visit schedules have been defined using the "
                f"offstudy model '{offstudy_model}'"
            )
        return visit_schedules

    def get_consent_model(
        self,
        visit_schedule_name: str,
        schedule_name: str,
        site: SingleSite | None = None,
    ) -> str:
        """Returns the consent model name specified on the schedule"""
        schedule = self.get_visit_schedule(visit_schedule_name).schedules.get(schedule_name)
        if isinstance(schedule.consent_model, (dict,)):
            # schedule returns a dict, get model name for this
            # site_id or country
            consent_model = schedule.consent_model.get(
                site.site_id
            ) or schedule.consent_model.get(site.country)
        else:
            # schedule returns a string
            consent_model = schedule.consent_model
        return consent_model

    def get_onschedule_model(self, visit_schedule_name: str, schedule_name: str) -> str:
        """Returns the onschedule model name"""
        schedule = self.get_visit_schedule(visit_schedule_name).schedules.get(schedule_name)
        return schedule.onschedule_model

    @staticmethod
    def get_offstudy_model() -> str:
        offstudy_models = []
        for visit_schedule in site_visit_schedules.get_visit_schedules().values():
            if visit_schedule.offstudy_model not in offstudy_models:
                offstudy_models.append(visit_schedule.offstudy_model)
        if len(offstudy_models) > 1:
            raise SiteVisitScheduleError(
                "More than one off study model defined. See visit schedules. "
                f"Got {offstudy_models}."
            )
        if len(offstudy_models) == 0:
            visit_schedule_names = [
                k for k, v in site_visit_schedules.get_visit_schedules().items()
            ]
            raise SiteVisitScheduleError(
                "No off study model defined in visit_schedule. "
                f"Got registered visit_schedules: {visit_schedule_names}."
            )
        return offstudy_models[0]

    @property
    def all_post_consent_models(self) -> dict[str, str]:
        """Returns a dictionary of models that require consent before save.

        {model_name1: consent_model_name, model_name2: consent_model_name, ...}
        """
        if not self._all_post_consent_models:
            models = {}
            for visit_schedule in self.visit_schedules.values():
                models.update(**visit_schedule.all_post_consent_models)
            self._all_post_consent_models = models
        return self._all_post_consent_models

    @staticmethod
    def to_model(model_cls: VisitScheduleModel) -> None:
        """Updates the VisitSchedule model with the current visit
        schedule, schedule and visits.

        Note: The VisitSchedule model is just for reference and does
        not replace information gathered from -this- class.

        Note: in addition to other attrs, such as visit code,
        timepoint must be unique across all schedules. Timepoint
        will not be updated for an existing record and may raise
        an integrity error. If you change a timepoint value within
        an existing schedule, you may need to change it manually via
        the database client.
        """
        model_cls.objects.update(active=False)
        for visit_schedule in site_visit_schedules.visit_schedules.values():
            for schedule in visit_schedule.schedules.values():
                for visit in schedule.visits.values():
                    opts = dict(
                        visit_schedule_name=visit_schedule.name,
                        schedule_name=schedule.name,
                        visit_code=visit.code,
                        visit_name=visit.name,
                        visit_title=visit.title,
                        timepoint=visit.timepoint,
                        active=True,
                    )
                    try:
                        obj = model_cls.objects.get(
                            visit_schedule_name=visit_schedule.name,
                            schedule_name=schedule.name,
                            timepoint=visit.timepoint,
                        )
                    except ObjectDoesNotExist:
                        model_cls.objects.create(**opts)
                    else:
                        for fld, value in opts.items():
                            setattr(obj, fld, value)
                        obj.save()

    def autodiscover(self, module_name=None, apps=None, verbose=None) -> None:
        """Autodiscovers classes in the visit_schedules.py file of
        any INSTALLED_APP.
        """
        self.loaded = True
        before_import_registry = None
        module_name = module_name or "visit_schedules"
        verbose = True if verbose is None else verbose
        if verbose:
            sys.stdout.write(f" * checking site for module '{module_name}' ...\n")
        for app in apps or django_apps.app_configs:
            try:
                mod = import_module(app)
                try:
                    before_import_registry = copy.copy(site_visit_schedules._registry)
                    import_module(f"{app}.{module_name}")
                    if verbose:
                        sys.stdout.write(f"   - registered visit schedule from '{app}'\n")
                except Exception as e:
                    if f"No module named '{app}.{module_name}'" not in str(e):
                        raise
                    site_visit_schedules._registry = before_import_registry
                    if module_has_submodule(mod, module_name):
                        raise
            except ModuleNotFoundError:
                pass


site_visit_schedules = SiteVisitSchedules()



from .crf import CrfRule, CrfRuleGroup, CrfRuleModelConflict
from .decorators import RegisterRuleGroupError, register
from .logic import Logic, RuleLogicError
from .metadata_rule_evaluator import MetadataRuleEvaluator
from .persistant_singleton_mixin import PersistantSingletonMixin
from .predicate import PF, NoValueError, P, PredicateError
from .requisition import (
    RequisitionRule,
    RequisitionRuleGroup,
    RequisitionRuleGroupMetaOptionsError,
)
from .rule import Rule, RuleError
from .rule_evaluator import RuleEvaluatorError, RuleEvaluatorRegisterSubjectError
from .rule_group import TargetModelConflict
from .rule_group_meta_options import RuleGroupMetaError
from .rule_group_metaclass import RuleGroupError
from .site import (
    SiteMetadataNoRulesError,
    SiteMetadataRulesAlreadyRegistered,
    site_metadata_rules,
)



from __future__ import annotations

from typing import TYPE_CHECKING

from ..formula import Formula
from .normal_data_model_cls import normal_data_model_cls

if TYPE_CHECKING:
    from ..models import ReferenceRangeCollection


def update_normal_data(
    reference_range_collection: ReferenceRangeCollection,
    normal_data: dict[str, list[Formula]] | None = None,
):
    normal_data_model_cls().objects.filter(
        reference_range_collection=reference_range_collection
    ).delete()
    for label, formulas in normal_data.items():
        for formula in formulas:
            opts = {k: v for k, v in formula.__dict__.items() if k != "gender"}
            for gender in formula.__dict__.get("gender"):
                normal_data_model_cls().objects.create(
                    reference_range_collection=reference_range_collection,
                    label=label,
                    gender=gender,
                    **opts,
                )



from __future__ import annotations

from django.db import models

from edc_model.models import HistoricalRecords
from edc_search.model_mixins import SearchSlugManager
from edc_sites.managers import CurrentSiteManager

from ..screening_identifier import ScreeningIdentifier
from .screening_fields_model_mixin import ScreeningFieldsModeMixin
from .screening_identifier_model_mixin import ScreeningIdentifierModelMixin
from .screening_methods_model_mixin import ScreeningMethodsModeMixin


class ScreeningManager(SearchSlugManager, models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, screening_identifier):
        return self.get(screening_identifier=screening_identifier)


class ScreeningModelMixin(
    ScreeningMethodsModeMixin,
    ScreeningIdentifierModelMixin,
    ScreeningFieldsModeMixin,
    models.Model,
):
    """You may wish to also include the `EligibilityModelMixin`
    in your declaration."""

    identifier_cls = ScreeningIdentifier

    # add `site` to your concrete model if asking for site confirmation
    # and on the screening form to make the field editable.
    # site = models.ForeignKey(Site, on_delete=models.PROTECT, null=True, related_name="+")

    objects = ScreeningManager()

    on_site = CurrentSiteManager()

    history = HistoricalRecords(inherit=True)

    class Meta(ScreeningIdentifierModelMixin.Meta):
        abstract = True
        indexes = ScreeningIdentifierModelMixin.Meta.indexes



from dateutil.relativedelta import relativedelta
from django.core.exceptions import ObjectDoesNotExist

from edc_visit_tracking.utils import get_previous_related_visit


def update_previous_refill_end_datetime(instance):
    """Update refill_end_datetime from previous visit relative to the
    refill_start_datetime of this visit.
    """
    if previous_visit := get_previous_related_visit(
        instance.related_visit, include_interim=True
    ):
        opts = {instance.__class__.related_visit_model_attr(): previous_visit}
        try:
            obj = instance.__class__.objects.get(**opts)
        except ObjectDoesNotExist:
            pass
        else:
            obj.refill_end_datetime = instance.refill_start_datetime - relativedelta(seconds=1)
            obj.save_base(update_fields=["refill_end_datetime"])


__all__ = ["update_previous_refill_end_datetime"]



from django.db import models
from django_crypto_fields.fields import EncryptedCharField

from edc_model.models import BaseUuidModel, HistoricalRecords


class Manager(models.Manager):
    use_in_migrations = True


class Assignment(BaseUuidModel):

    name = EncryptedCharField(help_text="word as in randomization list")

    display_name = EncryptedCharField(
        verbose_name="Formal label",
        null=True,
        blank=True,
        help_text="If not provided, defaults to 'assignment'",
    )

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return self.display_name

    def save(self, *args, **kwargs):
        self.assignment = self.name.lower()
        if not self.display_name:
            self.display_name = self.assignment.title()
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Assignment"
        verbose_name_plural = "Assignments"



from ..df_handlers import NonCrfDfHandler
from .tables_exporter import TablesExporter


class CsvExporterNoTables(Exception):
    pass


class CsvNonCrfTablesExporterError(Exception):
    pass


class CsvNonCrfTablesExporter(TablesExporter):
    """A class to export non-CRF tables for this app_label."""

    # a list of columns to NOT appear in any table, used to select the tables
    without_crf_columns = None
    crf_dialect_cls = NonCrfDfHandler

    def __init__(self, without_columns=None, without_crf_columns=None, **kwargs):
        without_crf_columns = without_crf_columns or self.without_crf_columns or []
        without_columns = without_columns or []
        without_columns.extend(without_crf_columns)
        super().__init__(without_columns=without_columns, **kwargs)



from django.utils.safestring import mark_safe

from .site_notifications import site_notifications


class NotificationModelAdminMixin:
    """Show some information about which notifications are
    linked to the form and of those which are the current
    user subscribed to.

    Requires to be declared together with `ModelAdminFormInstructionsMixin`
    from module `edc_model_admin`.
    """

    def get_notification_instructions(self, request=None):
        notifications = site_notifications.models.get(self.model._meta.label_lower)
        notification_instructions = None
        if notifications:
            notifications = [notification.display_name for notification in notifications]
            tooltip1 = mark_safe(", ".join(notifications))  # nosec B308 B703
            my_notifications = [
                n.display_name
                for n in request.user.userprofile.email_notifications.filter(
                    display_name__in=notifications
                )
            ]
            tooltip2 = mark_safe(", ".join(my_notifications))  # nosec B308 B703
            word = "notification is" if len(notifications) == 1 else "notifications are"
            notification_instructions = (
                f'<a href="#" title="{tooltip1}">{len(notifications)} '
                f"{word}</a> enabled for this form. "
                f'You are <a href="#" title="{tooltip2}">subscribed '
                f"to {len(my_notifications)}</a>. "
                f"See your user profile for more details."
            )
            return mark_safe(notification_instructions)  # nosec B308 B703
        return notification_instructions

    def get_add_instructions(self, extra_context, request=None):
        extra_context = super().get_add_instructions(extra_context)
        extra_context["notification_instructions"] = self.get_notification_instructions(
            request
        )
        return extra_context

    def get_change_instructions(self, extra_context, request=None):
        extra_context = super().get_add_instructions(extra_context)
        extra_context["notification_instructions"] = self.get_notification_instructions(
            request
        )
        return extra_context



from __future__ import annotations

from io import BytesIO
from tempfile import mkdtemp
from typing import TYPE_CHECKING

from django.core.handlers.wsgi import WSGIRequest
from django.db import models
from django.db.models import QuerySet
from pypdf import PdfWriter

from .numbered_canvas import NumberedCanvas

if TYPE_CHECKING:
    from .model_mixins import PdfReportModelMixin

    class Model(PdfReportModelMixin, models.Model): ...


mkdtemp()


def write_queryset_to_secure_pdf(
    queryset: QuerySet | None = None,
    password: str | None = None,
    request: WSGIRequest | None = None,
    **extra_context,
) -> BytesIO:
    """Merges one or more PDF buffers into a single secure PDF buffer
    with a password.

    Pass PDF buffer to FileResponse or write to file.
    """
    merger = PdfWriter()
    for model_obj in queryset:
        buffer = write_model_to_insecure_pdf(model_obj, request=request, **extra_context)
        merger.append(fileobj=buffer)
        buffer.close()
    merged_buffer = BytesIO()
    merger.encrypt(password, algorithm="AES-256")
    merger.write(merged_buffer)
    merged_buffer.seek(0)
    return merged_buffer


def write_model_to_insecure_pdf(
    model_obj: Model, request: WSGIRequest | None = None, **extra_context
) -> BytesIO:
    pdf_report = model_obj.pdf_report_cls(
        model_obj=model_obj, request=request, **extra_context
    )
    buffer = BytesIO()
    doctemplate = pdf_report.document_template(buffer, **pdf_report.page)
    story = pdf_report.get_report_story()
    doctemplate.build(
        story,
        onFirstPage=pdf_report.on_first_page,
        onLaterPages=pdf_report.on_later_pages,
        canvasmaker=NumberedCanvas,
    )
    buffer.seek(0)
    return buffer



from collections.abc import Iterable

from django import forms
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured, ObjectDoesNotExist
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from .auth_objects import RANDO_UNBLINDED


def get_unblinded_users() -> list[str]:
    unblinded_users = getattr(settings, "EDC_RANDOMIZATION_UNBLINDED_USERS", [])
    if not trial_is_blinded() and unblinded_users:
        raise ImproperlyConfigured(
            "Did not expect a list of unblinded users. Got settings."
            "EDC_RANDOMIZATION_BLINDED_TRIAL=False but "
            f"EDC_RANDOMIZATION_UNBLINDED_USERS={unblinded_users}"
        )
    return unblinded_users


def trial_is_blinded() -> bool:
    """Default is True"""
    blinded_trial = getattr(settings, "EDC_RANDOMIZATION_BLINDED_TRIAL", True)
    if blinded_trial is None:
        blinded_trial = True
    return blinded_trial


def user_is_blinded(username) -> bool:
    if blinded := trial_is_blinded():
        try:
            user = get_user_model().objects.get(
                username=username, is_staff=True, is_active=True
            )
        except ObjectDoesNotExist:
            blinded = True
        else:
            if user.username in get_unblinded_users():
                blinded = False
    return blinded


def user_is_blinded_from_request(request):
    if user_is_blinded(request.user.username) or (
        not user_is_blinded(request.user.username)
        and RANDO_UNBLINDED not in [g.name for g in request.user.groups.all()]
    ):
        return True
    return False


def raise_if_prohibited_from_unblinded_rando_group(username: str, groups: Iterable) -> None:
    """A user form validation to prevent adding an unlisted
    user to the RANDO_UNBLINDED group.

    See also edc_auth's UserForm.
    """
    if RANDO_UNBLINDED in [grp.name for grp in groups] and user_is_blinded(username):
        raise forms.ValidationError(
            {
                "groups": format_html(
                    "{}",
                    mark_safe(
                        "This user is not unblinded and may not added "
                        "to the <U>RANDO_UNBLINDED</U> group."
                    ),  # nosec B703 B308
                )
            }
        )



from django.conf import settings
from django.core.checks import CheckMessage, Error


def check_for_edc_model(app_configs, **kwargs) -> list[CheckMessage]:
    errors = []
    if not settings.USE_TZ:
        errors.append(Error("EDC requires settings.USE_TZ = True", id="settings.USE_TZ"))

    return errors



import os


def get_export_folder() -> str:
    from django.conf import settings

    if path := getattr(settings, "EDC_EXPORT_EXPORT_FOLDER", None):
        return os.path.expanduser(path)
    return os.path.join(settings.MEDIA_ROOT, "data_folder", "export")



from edc_visit_schedule.utils import raise_if_not_baseline


class ClinicalReviewBaselineFormValidatorMixin:
    def _clean(self):
        raise_if_not_baseline(self.cleaned_data.get("subject_visit"))
        super()._clean()



from django_audit_fields import audit_fieldset_tuple

from ..fieldsets import REQUIRES_CONSENT_FIELDS, requires_consent_fieldset_tuple


class RequiresConsentModelAdminMixin:

    def get_fieldsets(self, request, obj=None):
        fieldsets = super().get_fieldsets(request, obj=obj)
        fieldsets = list(fieldsets)
        for index, fieldset in enumerate(fieldsets):
            if fieldset == audit_fieldset_tuple:
                fieldsets.insert(index, requires_consent_fieldset_tuple)
                break
        return tuple(fieldsets)

    def get_readonly_fields(self, request, obj=None) -> tuple[str, ...]:
        readonly_fields = super().get_readonly_fields(request, obj=obj)
        return tuple(set(readonly_fields + REQUIRES_CONSENT_FIELDS))



from __future__ import annotations

from django.db import models

from edc_constants.choices import YES_NO_NA
from edc_constants.constants import CHOL, DM, HIV, HTN, NOT_APPLICABLE
from edc_dx import get_diagnosis_labels

default_prompts = {
    HIV.lower(): "Since last seen, has the patient tested <U>positive</U> for HIV infection?",
    DM: "Since last seen, has the patient been diagnosed with Diabetes?",
    HTN: "Since last seen, has the patient been diagnosed with Hypertension?",
    CHOL: "Since last seen, has the patient been diagnosed with High Cholesterol?",
}


def followup_review_model_mixin_factory(prompts: dict[str, str] | None = None):
    prompts = prompts or default_prompts

    class AbstractModel(models.Model):
        class Meta:
            abstract = True

    opts = {}
    for dx in get_diagnosis_labels():
        opts.update(
            {
                f"{dx}_dx": models.CharField(
                    verbose_name=prompts.get(dx),
                    max_length=15,
                    choices=YES_NO_NA,
                    default=NOT_APPLICABLE,
                )
            }
        )

    for name, fld_cls in opts.items():
        AbstractModel.add_to_class(name, fld_cls)

    return AbstractModel



from django.contrib import admin
from django_revision.modeladmin_mixin import ModelAdminRevisionMixin

from edc_model_admin.mixins import (
    ModelAdminFormAutoNumberMixin,
    ModelAdminFormInstructionsMixin,
    ModelAdminInstitutionMixin,
    ModelAdminNextUrlRedirectMixin,
    ModelAdminRedirectOnDeleteMixin,
    TemplatesModelAdminMixin,
)

from ..actions import print_test_label_sheet_action
from ..admin_site import edc_pylabels_admin
from ..forms import LabelConfigurationForm
from ..models import LabelConfiguration
from ..site_label_configs import site_label_configs


@admin.register(LabelConfiguration, site=edc_pylabels_admin)
class LabelConfigurationAdmin(
    TemplatesModelAdminMixin,
    ModelAdminNextUrlRedirectMixin,  # add
    ModelAdminFormInstructionsMixin,  # add
    ModelAdminFormAutoNumberMixin,
    ModelAdminRevisionMixin,  # add
    ModelAdminInstitutionMixin,  # add
    ModelAdminRedirectOnDeleteMixin,
    admin.ModelAdmin,
):
    show_object_tools = True
    actions = [print_test_label_sheet_action]
    form = LabelConfigurationForm

    instructions = (
        "This model links the label specification with a registered label configuration."
    )

    date_hierarchy = "created"

    fieldsets = ((None, {"fields": ("name", "label_specification", "requires_allocation")}),)

    list_display = ("name", "label_specification")

    list_filter = ("label_specification",)

    search_fields = ("name", "label_specification__name")

    def formfield_for_choice_field(self, db_field, request, **kwargs):
        if db_field.name == "name":
            kwargs["choices"] = [(k, k) for k in site_label_configs.all()]
        return super().formfield_for_choice_field(db_field, request, **kwargs)



from __future__ import annotations

from typing import Any

from django.db.models import QuerySet

from edc_constants.constants import NEW, OPEN
from edc_sites.site import sites

from ..models import ActionItem


class ActionItemViewMixin:
    def get_context_data(self, **kwargs) -> dict[str, Any]:
        kwargs.update(open_action_items=self.open_action_items)
        return super().get_context_data(**kwargs)

    @property
    def open_action_items(self) -> QuerySet[ActionItem]:
        """Returns a list of wrapped ActionItem instances
        where status is NEW or OPEN.
        """
        return ActionItem.objects.filter(
            subject_identifier=self.kwargs.get("subject_identifier"),
            action_type__show_on_dashboard=True,
            status__in=[NEW, OPEN],
            site_id__in=sites.get_site_ids_for_user(request=self.request),
        ).order_by("-report_datetime")



from django.db.models.fields import CharField
from django.utils.translation import gettext as _

from edc_constants.choices import DATE_ESTIMATED, DATE_ESTIMATED_NA
from edc_constants.constants import NOT_APPLICABLE


class IsDateEstimatedField(CharField):
    """field to question if date is estimated"""

    description = _("Custom field to question if date is estimated")

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("editable", True)
        kwargs.setdefault("max_length", 25)
        kwargs.setdefault("choices", DATE_ESTIMATED)
        kwargs.setdefault(
            "help_text",
            _(
                "If the exact date is not known, please indicate which "
                "part of the date is estimated."
            ),
        )
        CharField.__init__(self, *args, **kwargs)

    def get_internal_type(self):
        return "CharField"


class IsDateEstimatedFieldNa(CharField):
    """field to question if date is estimated"""

    description = _("Custom field to question if date is estimated")

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("editable", True)
        kwargs.setdefault("null", False)
        kwargs.setdefault("blank", False)
        kwargs.setdefault("max_length", 25)
        kwargs.setdefault("choices", DATE_ESTIMATED_NA)
        kwargs.setdefault("default", NOT_APPLICABLE)
        kwargs.setdefault(
            "help_text",
            _(
                "If the exact date is not known, please indicate which "
                "part of the date is estimated."
            ),
        )
        CharField.__init__(self, *args, **kwargs)

    def get_internal_type(self):
        return "CharField"



from django import forms
from django.core.exceptions import ObjectDoesNotExist

from edc_utils import formatted_datetime
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from ...utils import OffstudyError, raise_if_offstudy


class OffstudyCrfModelFormMixin:
    """ModelForm mixin for CRF Models.

    Must be declared with at least VisitScheduleCrfModelFormMixin
    """

    def clean(self):
        cleaned_data = super().clean()
        self.raise_if_offstudy_by_report_datetime()
        self.raise_if_offschedule_by_report_datetime()
        return cleaned_data

    def raise_if_offschedule_by_report_datetime(self):
        """Raises a ValidationError if the subject is offschedule before
        the report_datetime.
        """
        if self.get_subject_identifier() and self.report_datetime:
            visit_schedule = site_visit_schedules.get_visit_schedule(self.visit_schedule_name)
            schedule = visit_schedule.schedules.get(self.schedule_name)
            try:
                offschedule_obj = schedule.offschedule_model_cls.objects.get(
                    subject_identifier=self.get_subject_identifier(),
                    offschedule_datetime__lt=self.report_datetime,
                )
            except ObjectDoesNotExist:
                pass
            else:
                offschedule_datetime = formatted_datetime(offschedule_obj.offschedule_datetime)
                raise forms.ValidationError(
                    f"Subject was taken off schedule before this report datetime. "
                    f"Got subject_identifier='{self.get_subject_identifier()}', "
                    f"schedule='{visit_schedule.name}.{schedule.name}, '"
                    f"offschedule date='{offschedule_datetime}'."
                )

    def raise_if_offstudy_by_report_datetime(self):
        """Raises a ValidationError on a CRF if the subject is
        off study before the report_datetime.
        """
        if self.report_datetime:
            try:
                raise_if_offstudy(
                    source_obj=self.instance,
                    subject_identifier=self.get_subject_identifier(),
                    report_datetime=self.report_datetime,
                )
            except OffstudyError as e:
                raise forms.ValidationError(e)



from django.contrib import admin

from edc_model_admin.mixins import TemplatesModelAdminMixin


class ListModelAdminMixin(TemplatesModelAdminMixin, admin.ModelAdmin):
    ordering: tuple[str, ...] = ("display_index", "display_name")

    list_display: tuple[str, ...] = ("display_name", "name", "display_index")

    search_fields: tuple[str, ...] = ("display_name", "name")



from django.conf import settings


def get_default_navbar():
    """Returns the default navbar name.

    For example: inte_dashboard for project INTE.
    """
    return getattr(settings, "EDC_NAVBAR_DEFAULT", f"{settings.APP_NAME}_dashboard".lower())



from django.apps import apps as django_apps


def get_codenames(
    codenames: list[str],
    view_only_models: list[str] | None = None,
    exclude_models: list[str] | None = None,
) -> list[str]:
    """Return a list of all codenames for the role.

    See auths.py
    """
    view_only_models = view_only_models or []
    exclude_models = exclude_models or []
    for app_config in django_apps.get_app_configs():
        if app_config.name in ["edc_pharmacy"]:
            for model_cls in app_config.get_models():
                label_lower = model_cls._meta.label_lower
                app_name, model_name = label_lower.split(".")
                if label_lower in exclude_models:
                    continue
                if label_lower in view_only_models:
                    codenames.append(f"{app_name}.view_{model_name}")
                else:
                    for prefix in ["view_", "add_", "change_", "delete_"]:
                        codenames.append(f"{app_name}.{prefix}{model_name}")
    return codenames



from django.contrib import admin
from django.contrib.admin import SimpleListFilter
from django.template.loader import render_to_string
from django.urls import NoReverseMatch, reverse
from django.utils.translation import gettext
from django_audit_fields import audit_fieldset_tuple
from rangefilter.filters import DateRangeFilterBuilder

from edc_constants.constants import NO, YES
from edc_model_admin.history import SimpleHistoryAdmin
from edc_utils.date import to_local

from ...admin_site import edc_pharmacy_admin
from ...models import Allocation
from ...utils import get_related_or_none
from ..list_filters import AssignmentListFilter
from ..model_admin_mixin import ModelAdminMixin
from ..remove_fields_for_blinded_users import remove_fields_for_blinded_users


class TransferredFilter(SimpleListFilter):
    title = "Transferred"
    parameter_name = "transferred"

    def lookups(self, request, model_admin):
        return (YES, YES), (NO, NO)

    def queryset(self, request, queryset):
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.filter(stock__stocktransferitem__isnull=False)
            elif self.value() == NO:
                qs = queryset.filter(stock__stocktransferitem__isnull=True)
        return qs


class DispensedFilter(SimpleListFilter):
    title = "Dispensed"
    parameter_name = "dispensed"

    def lookups(self, request, model_admin):
        return (YES, YES), (NO, NO)

    def queryset(self, request, queryset):
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.filter(stock__dispenseitem__isnull=False)
            elif self.value() == NO:
                qs = queryset.filter(stock__dispenseitem_isnull=True)
        return qs


class HasStockFilter(SimpleListFilter):
    title = "Orphaned"
    parameter_name = "orphaned"

    def lookups(self, request, model_admin):
        return (YES, YES), (NO, NO)

    def queryset(self, request, queryset):
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.filter(stock__isnull=True)
            elif self.value() == NO:
                qs = queryset.filter(stock__isnull=False)
        return qs


@admin.register(Allocation, site=edc_pharmacy_admin)
class AllocationAdmin(ModelAdminMixin, SimpleHistoryAdmin):

    change_list_title = "Pharmacy: Allocations"
    change_form_title = "Pharmacy: Allocation"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    actions = ["delete_selected"]

    ordering = (
        "registered_subject__subject_identifier",
        "allocation_datetime",
    )

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "allocation_identifier",
                    "allocation_datetime",
                    "stock_request_item",
                    "registered_subject",
                    "allocated_by",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "identifier",
        "allocation_date",
        "transferred",
        "dispensed",
        "dashboard",
        "stock_changelist",
        "stock_request_changelist",
        "stock_product",
        "stock_container",
        "assignment",
        "allocated_by",
    )

    list_filter = (
        "stock__location",
        AssignmentListFilter,
        ("allocation_datetime", DateRangeFilterBuilder()),
        TransferredFilter,
        DispensedFilter,
        "allocated_by",
        HasStockFilter,
    )

    search_fields = (
        "id",
        "stock__code",
        "stock_request_item__id",
        "stock_request_item__stock_request__id",
        "registered_subject__subject_identifier",
    )

    readonly_fields = (
        "assignment",
        "allocation_identifier",
        "allocation_datetime",
        "registered_subject",
        "stock_request_item",
        "allocated_by",
    )

    def get_list_display(self, request):
        fields = super().get_list_display(request)
        fields = remove_fields_for_blinded_users(request, fields)
        return fields

    def get_list_filter(self, request):
        fields = super().get_list_filter(request)
        fields = remove_fields_for_blinded_users(request, fields)
        return fields

    def get_search_fields(self, request):
        fields = super().get_search_fields(request)
        fields = remove_fields_for_blinded_users(request, fields)
        return fields

    @admin.display(description="ALLOCATION #", ordering="allocation_identifier")
    def identifier(self, obj):
        return obj.allocation_identifier.split("-")[0]

    @admin.display(description="Allocation date", ordering="allocation_datetime")
    def allocation_date(self, obj):
        return to_local(obj.allocation_datetime).date()

    @admin.display(description="T", boolean=True)
    def transferred(self, obj):
        return True if obj.stock.stocktransferitem else False

    @admin.display(description="D", boolean=True)
    def dispensed(self, obj):
        if obj:
            return True if get_related_or_none(obj.stock, "dispenseitem") else False
        return None

    @admin.display(description="Product", ordering="stock__product")
    def stock_product(self, obj):
        return obj.stock.product.name

    @admin.display(description="Assignment", ordering="stock__product__assignment")
    def assignment(self, obj):
        return obj.stock.product.assignment

    @admin.display(description="Product", ordering="stock__product")
    def stock_container(self, obj):
        return obj.stock.container

    @admin.display(description="Request #")
    def stock_request_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stockrequest_changelist")
        url = f"{url}?q={obj.stock_request_item.stock_request.id}"
        context = dict(
            url=url,
            label=f"{obj.stock_request_item.stock_request.request_identifier}",
            title="Go to stock request",
        )
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="Stock #")
    def stock_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stock_changelist")
        url = f"{url}?q={obj.stock.code}"
        context = dict(url=url, label=f"{obj.stock.code}", title="Go to stock")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="Subject #", ordering="registered_subject__subject_identifier")
    def dashboard(self, obj=None, label=None):
        context = {}
        try:
            url = reverse(
                self.get_subject_dashboard_url_name(),
                kwargs={"subject_identifier": obj.registered_subject.subject_identifier},
            )
        except NoReverseMatch:
            url = None
        else:
            context = dict(
                title=gettext("Go to subject dashboard"),
                url=url,
                label=obj.registered_subject.subject_identifier,
            )
        if url:
            return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)
        return None



from __future__ import annotations

from django.contrib import admin, messages
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from edc_utils.celery import run_task_sync_or_async

from ...utils import process_repack_request_queryset


@admin.action(description="Process repack request")
def process_repack_request_action(modeladmin, request, queryset):
    """Action to process repack request.

    Redirects to process_repack_request.

    If celery is running, will run through the entire queryset otherwise
    just the first instance in the queryset.

    """
    repack_request_pks = [obj.pk for obj in queryset]
    task = run_task_sync_or_async(
        process_repack_request_queryset,
        repack_request_pks=repack_request_pks,
        username=request.user.username,
    )
    task_id = getattr(task, "id", None)
    queryset.update(task_id=task_id)

    # add messages for user
    messages.add_message(
        request,
        messages.SUCCESS,
        format_html(
            "{}",
            mark_safe(
                "Repack request submitted. <BR>Next, go to the ACTION menu below and "
                "(1)`Print labels`. Then (2) Label your stock "
                "containers with the printed labels. "
                "Once all stock is labelled, go to the ACTION menu below and "
                "(3) Select `Confirm repacked and labelled stock`. "
                f"Scan in the labels to CONFIRM the stock. ({task_id})"
            ),  # nosec B703, B308
        ),
    )
    if task_id:
        messages.add_message(
            request,
            messages.INFO,
            f"Task {task_id} is processing your repack requests.",
        )
    else:
        repack_request = queryset.first()
        messages.add_message(
            request,
            messages.INFO,
            (
                f"Processed only 1 of {queryset.count()} repack requests selected. "
                f"See {repack_request}."
            ),
        )
        messages.add_message(
            request,
            messages.ERROR,
            "Task workers not running. Contact data management.",
        )

    # redirect to changelist
    url = reverse("edc_pharmacy_admin:edc_pharmacy_repackrequest_changelist")
    if queryset.count() == 1:
        repack_request = queryset.first()
        url = f"{url}?q={repack_request.from_stock.code}"
    return HttpResponseRedirect(url)



from __future__ import annotations

import logging
from contextlib import suppress
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.core.management.color import color_style
from django.db import models
from django.utils.formats import localize

from edc_action_item.stubs import ActionItemStub
from edc_constants.constants import CLOSED, NEW, OPEN
from edc_model.constants import DEFAULT_BASE_FIELDS
from edc_sites.utils import valid_site_for_subject_or_raise

from .create_action_item import SingletonActionItemError, create_action_item
from .exceptions import ActionError
from .get_action_type import get_action_type
from .site_action_items import site_action_items

if TYPE_CHECKING:
    from .models import ActionItem, ActionType

logger = logging.getLogger(__name__)
style = color_style()

REFERENCE_MODEL_ERROR_CODE = "reference_model"
ACTION_CLASS_MISMATCH = (
    "Action class mismatch for given ActionItem. "
    "{action_item_action_cls} incorrectly passed "
    "to Action {action_cls}."
)
ACTION_GOT_UNLISTED_PARENT_ACTION_ITEM = (
    "Action class received an unlisted parent_action_item. "
    "Expected one of {parent_action_names}. "
    "Got '{parent_action_cls_name}' "
    "subject_identifier `{subject_identifier}`."
    "See Action {action_cls}."
)

ACTION_EXPECTS_RELATED_ACTION_ITEM = (
    "Action class expects a related_action_item. "
    "related_reference_fk_attr=`{related_reference_fk_attr}`. "
    "Got None for action based on action_item `{action_item}`. "
    "See Action {action_cls}."
)

ACTION_ITEM_DOES_NOT_EXIST = "{err} Got action_identifier={action_identifier}."
ACTION_ITEM_GET_OR_CREATE_FAILED = "Unable to get or create ActionItem. Got {opts}."
ACTION_NAME_IS_NONE = "Action name cannot be None. See {action_cls}."


class Action:
    """Main class to represent an Action.

    Note: that some attrs are used to create the ActionType. A
      change to one of these instance attr will only be reflected
      after post-migrate is run.
    """

    admin_site_name: str = None
    color_style: str = "danger"
    create_by_action: bool = None
    create_by_user: bool = None
    display_name: str = None
    help_text: str = None
    instructions = None
    name: str = None
    parent_action_names: tuple[str] | None = None
    enforce_parent_action_names: bool = True
    priority: bool = None
    reference_model: str = None
    related_reference_fk_attr: str = None
    related_reference_model: str = None
    show_link_to_add: bool = False
    show_link_to_changelist: bool = False
    show_on_dashboard: bool = None
    singleton: bool = False
    delete_with_reference_object: bool = False

    popover_title = "Action Item"

    action_item_model: str = "edc_action_item.actionitem"
    action_type_model: str = "edc_action_item.actiontype"
    next_actions: list[str] | None = None  # a list of Action classes which may include 'self'

    def __init__(
        self,
        action_item: ActionItemStub = None,
        reference_obj: models.Model = None,
        subject_identifier: str | None = None,
        action_identifier: str | None = None,
        parent_action_item: ActionItemStub | None = None,
        related_action_item: ActionItemStub | None = None,
        using: str | None = None,
        readonly: bool | None = None,
        skip_get_current_site: bool | None = None,
        site_id: int | None = None,
    ) -> None:
        self._action_item = action_item
        self._reference_obj = reference_obj

        self.parent_action_names = self.parent_action_names or ()

        self.messages: dict = {}

        self.action_identifier = action_identifier
        self.parent_action_item = parent_action_item
        self.related_action_item = related_action_item
        self.readonly = readonly
        self.subject_identifier = subject_identifier
        self.using = using
        self.skip_get_current_site = skip_get_current_site
        self.site_id = site_id

        if self.action_item.action_cls != self.__class__:
            raise ActionError(
                ACTION_CLASS_MISMATCH.format(
                    action_item_action_cls=self.action_item.action_cls,
                    action_cls=self.__class__,
                ),
                code="class type mismatch",
            )
        self.action_identifier = self.action_item.action_identifier
        self.linked_to_reference = self.action_item.linked_to_reference
        self.parent_action_item = self.action_item.parent_action_item
        self.related_action_item = self.action_item.related_action_item
        if not self.subject_identifier:
            self.subject_identifier = self.action_item.subject_identifier

        if (
            self.enforce_parent_action_names
            and self.parent_action_item
            and self.parent_action_item.action_cls.name not in self.parent_action_names
        ):
            raise ActionError(
                ACTION_GOT_UNLISTED_PARENT_ACTION_ITEM.format(
                    parent_action_names=self.parent_action_names,
                    parent_action_cls_name=self.parent_action_item.action_cls.name,
                    subject_identifier=self.subject_identifier,
                    action_cls=self.__class__,
                )
            )
        if not self.related_action_item and self.related_reference_fk_attr:
            raise ActionError(
                ACTION_EXPECTS_RELATED_ACTION_ITEM.format(
                    related_reference_fk_attr=self.related_reference_fk_attr,
                    action_item=self.action_item,
                    action_cls=self.__class__,
                )
            )

        if self.reference_obj and not self.readonly:
            self.close_and_create_next()

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"

    def __str__(self):
        return self.name

    def get_color_style(self):
        return self.color_style

    def get_display_name(self) -> str:
        return self.action_item.action_type.display_name

    def get_priority(self) -> bool:
        return self.priority

    def get_popover_title(self) -> str:
        return self.popover_title

    def get_status(self) -> str:
        return self.action_item.get_status_display()

    @property
    def reference_obj(self):
        """Returns the reference model instance or None.

        The reference model instance "should" exist if
        action item exists and is CLOSED. If not, re-open.
        """
        if not self._reference_obj and self.action_identifier:
            try:
                self._reference_obj = (
                    self.reference_model_cls()
                    .objects.using(self.using)
                    .get(action_identifier=self.action_identifier)
                )
            except ObjectDoesNotExist:
                if self.action_item and self.action_item.status == CLOSED:
                    self.action_item.status = OPEN
                    self.action_item.save(update_fields=["status"], using=self.using)
                    self.action_item.refresh_from_db()
        return self._reference_obj

    @property
    def action_item(self):
        """Returns an action_item model instance.

        If necessary, create.
        """
        if not self._action_item:
            if self.action_identifier:
                try:
                    self._action_item = (
                        self.action_item_model_cls()
                        .objects.using(self.using)
                        .get(action_identifier=self.action_identifier)
                    )
                except ObjectDoesNotExist as e:
                    raise ObjectDoesNotExist(
                        ACTION_ITEM_DOES_NOT_EXIST.format(
                            err=str(e), action_identifier=self.action_identifier
                        )
                    ) from e
            elif self.reference_obj:
                self._action_item = self.reference_obj.action_item
            else:
                opts = dict(
                    subject_identifier=self.subject_identifier,
                    action_type=get_action_type(self.__class__),
                    related_action_item=self.related_action_item,
                    status=NEW,
                )
                try:
                    self._action_item = (
                        self.action_item_model_cls().objects.using(self.using).get(**opts)
                    )
                except ObjectDoesNotExist:
                    # does not exist so create ...
                    self._create_new_action_item(**opts)
                except MultipleObjectsReturned:
                    self._action_item = (
                        self.action_item_model_cls()
                        .objects.using(self.using)
                        .filter(**opts)
                        .order_by("created")[0]
                    )
            if not self._action_item:
                opts = dict(
                    reference_obj=self.reference_obj,
                    subject_identifier=self.subject_identifier,
                    action_identifier=self.action_identifier,
                    parent_action_item=self.parent_action_item,
                    related_action_item=self.related_action_item,
                )
                raise ActionError(ACTION_ITEM_GET_OR_CREATE_FAILED.format(opts=opts))
        return self._action_item

    def _create_new_action_item(self, subject_identifier: str | None = None, **opts):
        """Create a new action item.

        Called only after checking.
        """
        valid_site_for_subject_or_raise(
            subject_identifier, skip_get_current_site=self.skip_get_current_site
        )

        try:
            self._action_item = create_action_item(
                self.__class__,
                subject_identifier=subject_identifier,
                parent_action_item=self.parent_action_item,
                priority=self.get_priority(),
                using=self.using,
                site_id=self.site_id,
                **opts,
            )
        except SingletonActionItemError:
            self._action_item = (
                self.action_item_model_cls()
                .objects.using(self.using)
                .get(
                    subject_identifier=self.subject_identifier,
                    action_type=get_action_type(self.__class__),
                )
            )

    @classmethod
    def action_item_model_cls(cls) -> type[ActionItem]:
        """Returns the ActionItem model class."""
        return django_apps.get_model(cls.action_item_model)

    @classmethod
    def action_type_model_cls(cls) -> type[ActionType]:
        """Returns the ActionType model class."""
        return django_apps.get_model(cls.action_type_model)

    @classmethod
    def get_reference_model(cls):
        """Returns the reference model label lower."""
        return cls.reference_model

    @classmethod
    def reference_model_cls(cls):
        """Returns the reference model class."""
        return django_apps.get_model(cls.get_reference_model())

    @classmethod
    def related_reference_model_cls(cls):
        """Returns the related reference model class"""
        return django_apps.get_model(cls.related_reference_model)

    @classmethod
    def as_dict(cls) -> dict:
        """Returns select class attrs as a dictionary."""
        dct = {k: v for k, v in cls.__dict__.items() if not k.startswith("_")}
        with suppress(AttributeError):
            dct.update(reference_model=cls.get_reference_model().lower())
        with suppress(AttributeError):
            dct.update(related_reference_model=cls.related_reference_model.lower())
        dct.update(
            name=cls.name,
            display_name=cls.display_name,
            show_on_dashboard=(
                True if cls.show_on_dashboard is None else cls.show_on_dashboard
            ),
            show_link_to_changelist=(
                True if cls.show_link_to_changelist is None else cls.show_link_to_changelist
            ),
            create_by_user=(True if cls.create_by_user is None else cls.create_by_user),
            create_by_action=(True if cls.create_by_action is None else cls.create_by_action),
            instructions=cls.instructions,
        )
        return dct

    def get_next_actions(self) -> list[str]:
        """Returns a list of action classes to be created
        again by this model if the first has been closed on post_save.
        """
        return self.next_actions or []

    def close_action_item_on_save(self) -> bool:
        """Returns True if action item for \'action_identifier\'
        is to be closed on post_save.

        Note: post_save watches the reference model instance.
        """
        return True

    def close_and_create_next(self) -> None:
        """Attempt to close the action item and
        create new ones, if required.
        """
        if self.reference_obj_has_changed:
            self.reopen_action_items()
        status = CLOSED if self.close_action_item_on_save() else OPEN
        if self.action_item.status != status:
            self.action_item.status = status
            self.action_item.save(using=self.using)
            self.action_item.refresh_from_db(using=self.using)
        if status == CLOSED:
            self.create_next_action_items()

    def create_next_action_items(self) -> None:
        """Creates any next action items if they do not
        already exist.
        """
        next_actions = list(set(self.get_next_actions()))
        for action_name in next_actions:
            action_cls = (
                self.__class__ if action_name == "self" else site_action_items.get(action_name)
            )
            action_type = get_action_type(action_cls)
            if action_type.related_reference_model:
                related_action_item = self.action_item.related_action_item or self.action_item
            else:
                related_action_item = None
            action_cls(
                subject_identifier=self.subject_identifier,
                parent_action_item=self.action_item,
                related_action_item=related_action_item,
                using=self.using,
                skip_get_current_site=self.skip_get_current_site,
                site_id=self.site_id,
            )

    def reopen_action_item_on_change(self):
        """May be overriden."""
        return True

    def reopen_action_items(self):
        """Reopens the action_item and child action items for this
        reference object if reference object was changed since
        the last save.
        """
        for action_item in (
            self.action_item_model_cls()
            .objects.using(self.using)
            .filter(
                (
                    models.Q(action_identifier=self.reference_obj.action_identifier)
                    | models.Q(
                        parent_action_item__action_identifier=(
                            self.reference_obj.action_identifier
                        )
                    )
                    | models.Q(related_action_item=self.reference_obj.action_item)
                ),
                status=CLOSED,
            )
        ):
            if self.reopen_action_item_on_change():
                action_item.status = OPEN
                action_item.save(using=self.using)
                self.messages.update(
                    {
                        action_item: (
                            f"{self.reference_obj._meta.verbose_name.title()} "
                            f"{self.reference_obj} was changed on "
                            f"{localize(self.reference_obj.modified)} "
                            f"({settings.TIME_ZONE})"
                        )
                    }
                )
                self.action_item.refresh_from_db()

    @property
    def reference_obj_has_changed(self):
        """Returns True if the reference object has changed
        since the last save.

        We get here during a post_save but before history

        Reviews the objects "history" (historical) instances.
        """
        changed_message = {}
        try:
            # should this be index 1 or 0? Was 1, changed to 0, the
            # most recent saved history. But how could self.reference_obj
            # be changed??
            history = (
                self.reference_obj.history.using(self.using).all().order_by("-history_date")[0]
            )
        except IndexError:
            pass
        except AttributeError:
            # suppressed here but is reviewed in system checks
            pass
        else:
            field_names = [
                field.name
                for field in self.reference_obj._meta.get_fields()
                if field.name not in DEFAULT_BASE_FIELDS
            ]
            for field_name in field_names:
                try:
                    if getattr(history, field_name) != getattr(self.reference_obj, field_name):
                        changed_message.update(
                            {field_name: getattr(self.reference_obj, field_name)}
                        )
                except AttributeError:
                    pass
        return changed_message

    def append_to_next_if_required(
        self, next_actions=None, action_cls=None, action_name=None, required=None
    ):
        """Returns next actions where the given action_cls.name is
        appended if required.

        `required` can be anything that evaluates to a boolean.

        Will not append if the ActionItem for the next action
        already exists.
        """
        with suppress(AttributeError):
            action_name = action_cls.name
        if not action_name:
            raise ActionError(ACTION_NAME_IS_NONE.format(action_cls=self))
        next_actions = next_actions or []
        required = True if required is None else required
        opts = dict(
            subject_identifier=self.subject_identifier,
            parent_action_item__action_identifier=self.action_identifier,
            action_type__name=action_name,
        )
        try:
            next_action_item = (
                self.action_item_model_cls().objects.using(self.using).get(**opts)
            )
        except ObjectDoesNotExist:
            next_action_item = None
        except MultipleObjectsReturned:
            # suggests the action item sequence is broken
            logger.warning(
                style.ERROR(
                    f"skipping 'append_to_next_if_required' for "
                    f"{self.action_identifier} next actions. "
                    f"You may wish to review this action item.\n"
                )
            )
            next_action_item = None
        if not next_action_item and required:
            next_actions.append(action_name)
        return next_actions

    def delete_children_if_new(self, parent_action_item=None):
        """Deletes the action item instance where status
        is NEW, use with caution.

        Since some actions are created by an event, this method
        could mess up the state.
        """
        index = 0
        opts = dict(
            subject_identifier=self.subject_identifier,
            parent_action_item=parent_action_item,
            status=NEW,
        )
        for index, obj in enumerate(  # noqa: B007
            self.action_item_model_cls().objects.using(self.using).filter(**opts)
        ):
            obj.delete(using=self.using)
        return index



from __future__ import annotations

from typing import TYPE_CHECKING

from django import template
from django.core.exceptions import ObjectDoesNotExist

from edc_metadata.constants import KEYED, REQUIRED
from edc_metadata.models import CrfMetadata, RequisitionMetadata

from ..site_prn_forms import site_prn_forms

if TYPE_CHECKING:
    from edc_appointment.models import Appointment
register = template.Library()

CRF = "CRF"
REQUISITION = "Requisition"


@register.inclusion_tag("edc_prn/list_prns.html")
def prn_list_items(subject_identifier, **kwargs):
    prn_forms = []
    for prn in site_prn_forms:
        if prn.get_show_on_dashboard(subject_identifier=subject_identifier):
            prn_forms.append(prn)
    if prn_forms:
        prn_forms = sorted(prn_forms, key=lambda x: x.verbose_name)
    return dict(prn_forms=prn_forms, subject_identifier=subject_identifier)


@register.inclusion_tag("edc_prn/add_prn_popover.html")
def add_prn_crf_popover(appointment: Appointment, subject_dashboard_url: str):
    prn_forms = []
    for crf in appointment.visits.get(appointment.visit_code).crfs_prn:
        if not CrfMetadata.objects.filter(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
            visit_code_sequence=appointment.visit_code_sequence,
            model=crf.model,
            entry_status__in=[REQUIRED, KEYED],
        ).exists():
            subject_visit_id = getattr(appointment.related_visit, "pk", None)
            crf.add_url = crf.model_cls().get_absolute_url()
            crf.related_visit_model_attr = crf.model_cls.related_visit_model_attr()
            crf.related_visit = str(subject_visit_id) if subject_visit_id else None
            prn_forms.append(crf)
    if prn_forms:
        prn_forms = sorted(prn_forms, key=lambda x: x.verbose_name)
    return dict(
        label=CRF,
        CRF=CRF,
        REQUISITION=REQUISITION,
        prn_forms=prn_forms,
        appointment_pk=str(appointment.pk),
        subject_identifier=appointment.subject_identifier,
        subject_dashboard_url=subject_dashboard_url,
    )


@register.inclusion_tag("edc_prn/add_prn_popover.html")
def add_prn_requisition_popover(appointment: Appointment, subject_dashboard_url):
    prn_forms = []
    for requisition in appointment.visits.get(appointment.visit_code).requisitions_prn:
        if not RequisitionMetadata.objects.filter(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
            visit_code_sequence=appointment.visit_code_sequence,
            model=requisition.model,
            panel_name=requisition.panel.name,
            entry_status__in=[REQUIRED, KEYED],
        ).exists():
            requisition.add_url = requisition.model_cls().get_absolute_url()
            requisition.related_visit_model_attr = (
                requisition.model_cls.related_visit_model_attr()
            )
            requisition.related_visit = str(getattr(appointment.related_visit, "id", ""))
            try:
                panel_id = requisition.model_cls.panel.field.remote_field.model.objects.get(
                    name=requisition.panel.name
                ).id
            except ObjectDoesNotExist:
                requisition.panel.id = None
                requisition.panel.pk = None
            else:
                requisition.panel.id = panel_id
                requisition.panel.pk = panel_id
            prn_forms.append(requisition)
    if prn_forms:
        prn_forms = sorted(prn_forms, key=lambda x: x.verbose_name)
    return dict(
        label=REQUISITION,
        CRF=CRF,
        REQUISITION=REQUISITION,
        prn_forms=prn_forms,
        appointment_pk=str(appointment.pk),
        subject_identifier=appointment.subject_identifier,
        subject_dashboard_url=subject_dashboard_url,
    )



from .model_exporter import (
    ModelExporter,
    ModelExporterError,
    ModelExporterInvalidLookup,
    ModelExporterUnknownField,
)
from .object_history_helpers import ObjectHistoryCreator
from .plan_exporter import PlanExporter
from .value_getter import ValueGetter, ValueGetterInvalidLookup, ValueGetterUnknownField



from __future__ import annotations

import re
from dataclasses import KW_ONLY, dataclass, field

from edc_constants.constants import FEMALE, MALE

from .adult_age_options import adult_age_options

__all__ = ["Formula", "FormulaError", "dummy_formula", "formula"]

from .exceptions import FormulaError


def clean_and_validate_phrase(phrase) -> str:
    phrase_pattern = (
        r"^(([\d+\.]*\d+)?\*?(LLN|ULN)?(<|<=)?(?!\s*=\s*))+x(?!\s*=\s*)"
        r"((<|<=)?([\d+\.]*\d+)?\*?(LLN|ULN)?)?$"
    )
    phrase = phrase.replace(" ", "")
    if not re.match(phrase_pattern, phrase):
        raise FormulaError(
            f"Invalid. Got {phrase}. Expected, e.g, 11<x<22, "
            "11<=x<22, 11<x<=22, 11<x, 11<=x, x<22, x<=22, etc."
        )
    return phrase


@dataclass()
class Formula:
    phrase: str
    _ = KW_ONLY
    units: str = None
    fasting: bool | None = field(default=False)
    gender: str | list[str] = field(default="")
    age_lower: str = field(default="")
    age_upper: str = field(default="")
    age_units: str = field(default="")
    lower_inclusive: bool = field(default=False)
    upper_inclusive: bool = field(default=False)
    age_lower_inclusive: bool | None = None
    age_upper_inclusive: bool | None = None
    grade: int | None = None
    lln: str | None = field(default="", init=False, repr=False)
    uln: str | None = field(default="", init=False, repr=False)
    lower_operator: str = field(default="", init=False, repr=False)
    upper_operator: str = field(default="", init=False, repr=False)
    age_lower_operator: str = field(default="", init=False, repr=False)
    age_upper_operator: str = field(default="", init=False, repr=False)

    def __post_init__(self):
        self.phrase = clean_and_validate_phrase(self.phrase)
        lower_str, upper_str = self.phrase.split("x")
        for label, _str in {"lower": lower_str, "upper": upper_str}.items():
            if "*" in _str and "LLN" not in _str and "ULN" not in _str:
                raise FormulaError(
                    f"Invalid {label} limit normal in formula. Expected one of LLN or ULN. "
                    f"Got `{self.phrase}`"
                )
        self.lower, self.lower_inclusive, self.lln = self.parse_fragment(lower_str)
        self.upper, self.upper_inclusive, self.uln = self.parse_fragment(upper_str)
        self.gender = [self.gender] if isinstance(self.gender, str) else self.gender
        self.lower_operator = "" if not self.lower else "<=" if self.lower_inclusive else "<"
        self.upper_operator = "" if not self.upper else "<=" if self.upper_inclusive else "<"
        self.age_lower_operator = (
            "" if not self.age_lower else "<=" if self.age_lower_inclusive else "<"
        )
        self.age_upper_operator = (
            "" if not self.age_upper else "<=" if self.age_upper_inclusive else "<"
        )
        valid_grades = [0, 1, 2, 3, 4, 5]
        if self.grade and self.grade not in valid_grades:
            raise FormulaError(
                f"Invalid grade. Expected one of {valid_grades}. Got `{self.grade}`"
            )

    def __str__(self) -> str:
        return self.description

    @property
    def description(self) -> str:
        try:
            fasting = self.fasting
        except KeyError:
            fasting_str = ""
        else:
            fasting_str = "Fasting " if fasting else ""
        grade = ""
        if self.grade is not None:
            grade = f"GRADE{self.grade} "
        return (
            f"{self.lower or ''}{self.lower_operator or ''}x{self.upper_operator or ''}"
            f"{self.upper or ''} "
            f"{self.units or 'units'} {fasting_str}{grade}{','.join(self.gender)} "
            f"{self.age_description}".rstrip()
        )

    @property
    def age_description(self) -> str:
        return (
            ""
            if not self.age_lower and not self.age_upper
            else (
                f"{self.age_lower or ''}{self.age_lower_operator}"
                f"AGE{self.age_upper_operator}{self.age_upper or ''}"
            )
        )

    @staticmethod
    def parse_fragment(fragment: str) -> tuple[float, bool, str | None]:
        limit_normal: str = ""
        inclusive = "=" in fragment
        try:
            value = float(
                fragment.replace("<", "")
                .replace("=", "")
                .replace("*LLN", "")
                .replace("*ULN", "")
            )
        except ValueError:
            value = None
        if "*LLN" in fragment:
            limit_normal = "*LLN"
        elif "*ULN" in fragment:
            limit_normal = "*ULN"
        return value, inclusive, limit_normal


def formula(phrase, **kwargs):
    return Formula(phrase, **kwargs).description


def dummy_formula(phrase: str, units: str) -> Formula:
    return Formula(
        phrase or "0<=x",
        units=units,
        gender=[MALE, FEMALE],
        **adult_age_options,
    )
